from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np

import dcmri.ui as ui
import dcmri.lib as lib
import dcmri.sig as sig
import dcmri.utils as utils
import dcmri.pk_aorta as pk_aorta
import dcmri.pk as pk
import dcmri.liver as liver


class AortaPortalLiver(ui.Model):
    """Joint model for aorta, portal vein and liver signals.

    This model uses a whole-body model to simultaneously predict signals in 
    aorta, portal vein and liver.  

    For more detail on the whole-body model, see :ref:`whole-body-tissues`. 
    For more detail on the liver model, see :ref:`liver-tissues`. 

    Args:
        kinetics (str, optional): Tracer-kinetic liver model. See table 
          :ref:`table-liver-models` for options - only dual-inlet models 
          are allowed. Defaults to '2I-EC'.
        stationary (str, optional): For intracellular tracers - stationarity 
          regime of the hepatocytes. The options are 'UE', 'E', 'U' or None. 
          For more detail see :ref:`liver-tissues`. Defaults to 'UE'.
        sequence (str, optional): imaging sequence. Possible values are 'SS'
          and 'SSI' (steady-state with aortic inflow correction). Defaults 
          to 'SS'.
        free (dict, optional): Dictionary with free parameters and their
          bounds. If not provided, a default set of free parameters is used.
          Defaults to None.
        params (dict, optional): values for the parameters of the tissue,
          specified as keyword parameters. Defaults are used for any that are
          not provided. See tables :ref:`AortaLiver-parameters` and
          :ref:`AortaLiver-defaults` for a list of parameters and their
          default values.

    See Also:
        `AortaLiver`

    Example:

        Use the model to reconstruct concentrations from experimentally 
        derived signals.

    .. plot::
        :include-source:
        :context: close-figs

        >>> import matplotlib.pyplot as plt
        >>> import dcmri as dc

        Use `fake_tissue` to generate synthetic test data from 
        experimentally-derived concentrations:

        Use `fake_liver` to generate synthetic test data:

        >>> time, aif, vif, roi, _ = dc.fake_liver(sequence='SSI')

        Since this model generates 3 time curves, the x- and y-data are 
        tuples:

        >>> xdata, ydata = (time, time, time), (aif, vif, roi)

        Build an aorta-portal-liver model and parameters to match the 
        conditions of the fake liver data:

        >>> model = dc.AortaPortalLiver(
        ...     kinetics = '2I-IC',
        ...     sequence = 'SSI',
        ...     dt = 0.5,
        ...     tmax = 180,
        ...     weight = 70,
        ...     agent = 'gadoxetate',
        ...     dose = 0.2,
        ...     rate = 3,
        ...     field_strength = 3.0,
        ...     t0 = 10,
        ...     TR = 0.005,
        ...     FA = 15,
        ...     TS = 0.5,
        ... )

        Train the model on the data:

        >>> model.train(xdata, ydata, xtol=1e-3)

        Plot the reconstructed signals and concentrations and compare 
        against the experimentally derived data:

        >>> model.plot(xdata, ydata)

        We can also have a look at the model parameters after training:

        >>> model.print_params(round_to=3)
        --------------------------------
        Free parameters with their stdev
        --------------------------------
        First bolus arrival time (BAT): 14.616 (1.1) sec
        Cardiac output (CO): 100.09 (2.736) mL/sec
        Heart-lung mean transit time (Thl): 14.402 (1.375) sec
        Heart-lung dispersion (Dhl): 0.391 (0.013) 
        Organs blood mean transit time (To): 27.811 (5.291) sec
        Organs extraction fraction (Eo): 0.29 (0.105)
        Organs extravascular mean transit time (Toe): 70.621 (102.614) sec
        Body extraction fraction (Eb): 0.013 (0.23)
        Aorta inflow time (TF): 0.409 (0.014) sec
        Liver extracellular volume fraction (ve): 0.479 (0.112) mL/cm3
        Liver plasma flow (Fp): 0.018 (0.001) mL/sec/cm3
        Arterial flow fraction (fa): 0.087 (0.074)
        Arterial transit time (Ta): 2.398 (1.356) sec
        Hepatocellular uptake rate (khe): 0.006 (0.003) mL/sec/cm3
        Hepatocellular mean transit time (Th): 683.604 (2554.75) sec
        Gut mean transit time (Tg): 10.782 (0.614) sec
        Gut dispersion (Dg): 0.893 (0.07)
        ----------------------------
        Fixed and derived parameters
        ----------------------------
        Hematocrit (H): 0.45
        Arterial venous blood flow (Fa): 0.002 mL/sec/cm3
        Portal venous blood flow (Fv): 0.016 mL/sec/cm3
        Extracellular mean transit time (Te): 27.216 sec
        Biliary tissue excretion rate (Kbh): 0.001 mL/sec/cm3
        Hepatocellular tissue uptake rate (Khe): 0.012 mL/sec/cm3
        Biliary excretion rate (kbh): 0.001 mL/sec/cm3

    Notes:

        Table :ref:`AortaPortalLiver-parameters` lists the parameters that are 
        relevant in each regime. Table :ref:`AortaPortalLiver-defaults` list 
        all possible parameters and their default settings. 

        .. _AortaPortalLiver-parameters:
        .. list-table:: **Aorta-Liver parameters**
            :widths: 20 30 30
            :header-rows: 1

            * - Parameters
              - When to use
              - Further detail
            * - dt, tmax
              - Always
              - Time axis for forward model
            * - t0, dose_tolerance
              - Always
              - For estimating baseline signal
            * - field_strength, weight, agent, dose, rate
              - Always
              - Injection protocol
            * - R10a, R10l, S0a, S0v, S0l 
              - Always
              - Precontrast R1 (:ref:`relaxation-params`) and 
                S0 (:ref:`params-per-sequence`)for aorta and liver 
            * - FA, TR, TS
              - Always
              - :ref:`params-per-sequence`
            * - TF
              - If **sequence** is 'SSI'
              - To model aorta inflow effects
            * - BAT, CO, Thl, Dhl, To, Eo, Tie, Eb
              - Always
              - :ref:`whole-body-tissues`
            * - Tg , Dg
              - Always
              - Gut dispersion
            * - H, ve, Fp, fa, Ta, Tg, khe, khe_i, kh_f, Th, Th_i, Th_f.
              - Depends on **kinetics** and **stationary**
              - :ref:`table-liver-models`

        .. _AortaPortalLiver-defaults:
        .. list-table:: **Aorta-Liver parameter defaults**
            :widths: 5 10 10 10 10
            :header-rows: 1

            * - Parameter
              - Type
              - Value
              - Bounds
              - Free/Fixed
            * - 
              - **Simulation**
              -
              - 
              - 
            * - dt
              - Simulation
              - 0.5
              - [0, inf]
              - Fixed
            * - tmax
              - Simulation
              - 120
              - [0, inf]
              - Fixed
            * - dose_tolerance
              - Simulation
              - 0.1
              - [0, 1]
              - Fixed
            * - 
              - **Injection**
              -
              - 
              - 
            * - field_strength
              - Injection
              - 3
              - [0, inf]
              - Fixed
            * - weight
              - Injection
              - 70
              - [0, inf]
              - Fixed
            * - agent
              - Injection
              - 'gadoxetate'
              - None
              - Fixed
            * - dose
              - Injection
              - 0.0125
              - [0, inf]
              - Fixed
            * - rate
              - Injection
              - 1
              - [0, inf]
              - Fixed
            * - 
              - **Signal**
              -
              - 
              - 
            * - R10a
              - Signal
              - 0.7
              - [0, inf]
              - Fixed
            * - R10l
              - Signal
              - 0.7
              - [0, inf]
              - Fixed
            * - S0a
              - Signal
              - 1
              - [0, inf]
              - Free
            * - S0v
              - Signal
              - 1
              - [0, inf]
              - Free
            * - S0l
              - Signal
              - 1
              - [0, inf]
              - Free
            * - 
              - **Sequence**
              -
              - 
              - 
            * - FA
              - Sequence
              - 15
              - [0, inf]
              - Fixed
            * - S0
              - Sequence
              - 1
              - [0, inf]
              - Fixed
            * - TF
              - Sequence
              - 0.5
              - [0, 10]
              - Free
            * - TR
              - Sequence
              - 0.005
              - [0, inf]
              - Fixed
            * - TS
              - Sequence
              - 0
              - [0, inf]
              - Fixed
            * - 
              - **Whole body**
              -
              - 
              - 
            * - BAT
              - Whole body
              - 1200
              - [0, inf]
              - Free
            * - CO
              - Whole body
              - 100
              - [0, inf]
              - Free
            * - Thl
              - Whole body
              - 10
              - [0, 30]
              - Free
            * - Dhl
              - Whole body
              - 0.2
              - [0.05, 0.95]
              - Free
            * - To
              - Whole body
              - 20
              - [0, 60]
              - Free
            * - Eo
              - Whole body
              - 0.15
              - [0, 0.5]
              - Free
            * - Toe
              - Whole body
              - 120
              - [0, 800]
              - Free
            * - Eb
              - Whole body
              - 0.05
              - [0.01, 0.15]
              - Free
            * - 
              - **Portal vein**
              -
              - 
              - 
            * - Tg
              - Kinetic
              - 15
              - [0.1, 60]
              - Free
            * - Dg
              - Kinetic
              - 0.85
              - [0, 1]
              - Free
            * - uv
              - Kinetic
              - 1
              - [0, 1]
              - Free
            * - 
              - **Liver**
              -
              - 
              - 
            * - H
              - Kinetic
              - 0.45
              - [0, 1]
              - Fixed
            * - ve
              - Kinetic
              - 0.3
              - [0.01, 0.6]
              - Free
            * - Fp
              - Kinetic
              - 0.01
              - [0, 0.1]
              - Free
            * - fa
              - Kinetic
              - 0.2
              - [0, 0.1]
              - Free
            * - Ta
              - Kinetic
              - 0.5
              - [0, 3]
              - Free
            * - khe
              - Kinetic
              - 0.003
              - [0, 0.1]
              - Free
            * - khe_i
              - Kinetic
              - 0.003
              - [0, 0.1]
              - Free
            * - khe_f
              - Kinetic
              - 0.003
              - [0, 0.1]
              - Free
            * - Th
              - Kinetic
              - 1800
              - [600, 36000]
              - Free
            * - Th_i
              - Kinetic
              - 1800
              - [600, 36000]
              - Free
            * - Th_f
              - Kinetic
              - 1800
              - [600, 36000]
              - Free
            * - vol
              - Kinetic
              - 1000
              - [0, 10000]
              - Free

    """

    def __init__(self, kinetics='2I-EC', stationary='UE', 
                 sequence='SS', free=None, **params):

        # Configuration
        self.organs = '2cxm' 
        self.kinetics = kinetics 
        self.sequence = sequence 
        self.stationary = stationary

        self._check_config()
        self._set_defaults(free=free, **params)

        # Internal flags
        self._predict = None

    def _check_config(self):
        if self.sequence not in ['SS', 'SSI']:
            raise ValueError(
                'Sequence ' + str(self.sequence) + ' is not available.')
        if self.kinetics[0] != '2':
            raise ValueError('Only dual-inlet models are allowed.')
        liver.params_liver(self.kinetics, self.stationary)

    def _params(self):
        return (PARAMS | PARAMS_WHOLE_BODY | PARAMS_SEQUENCE 
                | PARAMS_LIVER | PARAMS_PORTAL | PARAMS_DERIVED)
    
    def _model_pars(self):
        pars_sequence = {
            'SS': ['FA', 'TR', 'TS'], 
            'SSI': ['TF', 'FA', 'TR', 'TS'], 
        }
        pars = list(PARAMS.keys()) 
        pars += list(PARAMS_WHOLE_BODY.keys())
        pars += pars_sequence[self.sequence]
        pars += liver.params_liver(self.kinetics, self.stationary)
        pars += list(PARAMS_PORTAL.keys())
        pars += ['vol']
        return pars


    def _par_values(self, kin=False, export=False, seq=None):

        if seq is not None:
            pars = {
                'SS': ['FA', 'TR'], 
                'SSI': ['TF', 'FA', 'TR'], 
            }
            pars = pars[self.sequence]
            return {par: getattr(self, par) for par in pars}
        
        if kin:
            pars = liver.params_liver(self.kinetics, self.stationary)
            return {par: getattr(self, par) for par in pars}
        
        if export:
            pars = self._par_values()
            all = self._model_pars()
            p1 = liver.params_liver(self.kinetics, self.stationary)
            p2 = list(PARAMS_WHOLE_BODY.keys())
            p3 = list(PARAMS_DERIVED.keys())
            p4 = list(PARAMS_PORTAL.keys())
            retain = p1 + p2 + p3 + p4 + ['TF']
            discard = set(all) - set(retain)
            return {p: pars[p] for p in pars if p not in discard}
        
        pars = self._model_pars()
        p = {par: getattr(self, par) for par in pars}

        try:
            p['Fa'] = p['Fp']*p['fa']
        except KeyError:
            pass
        try:
            p['Fv'] = p['Fp']*(1-p['fa'])
        except KeyError:
            pass
        try:
            p['Te'] = _div(p['ve'], p['Fp'])
        except KeyError:
            pass
        try:
            p['Th'] = np.mean([p['Th_i'], p['Th_f']])
        except KeyError:
            pass
        try:
            p['khe'] = np.mean([p['khe_i'], p['khe_f']])
        except KeyError:
            pass
        try:
            p['Kbh'] = _div(1, p['Th'])
        except KeyError:
            pass
        try:
            p['Khe'] = _div(p['khe'], p['ve'])
        except KeyError:
            pass
        try:
            p['kbh'] = _div(1-p['ve'], p['Th'])
        except KeyError:
            pass
        try:
            p['kbh_i'] = _div(1-p['ve'], p['Th_i'])
        except KeyError:
            pass
        try:
            p['kbh_f'] = _div(1-p['ve'], p['Th_f'])
        except KeyError:
            pass
        try:
            p['E'] = p['khe']/(p['khe']+p['Fp'])
        except KeyError:
            try:
                p['E'] = p['khe']/(p['khe']+self.Fp)
            except:
                pass
        try:
            p['Ktrans'] = (1-p['E'])*p['khe']
        except KeyError:
            pass        

        if p['vol'] is not None:
            try:
                p['CL'] = p['khe']*p['vol']
            except KeyError:
                pass
            
        return p


    def _conc_aorta(self) -> np.ndarray:
        if self.organs == 'comp':
            organs = ['comp', (self.To,)]
        else:
            organs = ['2cxm', ([self.To, self.Toe], self.Eo)]
        self.t = np.arange(0, self.tmax, self.dt)
        conc = lib.ca_conc(self.agent)
        Ji = lib.ca_injection(
            self.t, self.weight, conc, self.dose, self.rate, self.BAT)
        Jb = pk_aorta.flux_aorta(
            Ji, E=self.Eb, dt=self.dt, tol=self.dose_tolerance,
            heartlung = ['pfcomp', (self.Thl, self.Dhl)],
            organs = organs)
        self.ca = Jb/self.CO
        return self.t, self.ca

    def _relax_aorta(self):
        t, cb = self._conc_aorta()
        rb = lib.relaxivity(self.field_strength, 'blood', self.agent)
        return t, self.R10a + rb*cb

    def _predict_aorta(self, xdata: np.ndarray) -> np.ndarray:
        self.tmax = max(xdata)+self.dt
        if self.TS is not None:
            self.tmax += self.TS
        t, R1b = self._relax_aorta()
        if self.sequence == 'SSI':
            signal = sig.signal_spgr(
                self.S0a, R1b, self.TF, self.TR, self.FA, n0=1)
        else:
            signal = sig.signal_ss(self.S0a, R1b, self.TR, self.FA)
        return utils.sample(xdata, t, signal, self.TS)
    
    def _conc_portal(self):
        self.cv = pk.flux_chain(self.ca, self.Tg, self.Dg, dt=self.dt)
        return self.cv
    
    def _relax_portal(self):
        rb = lib.relaxivity(self.field_strength, 'blood', self.agent)
        cv = self._conc_portal()
        return self.R10a + rb*self.uv*cv
    
    def _predict_portal(self, xdata):
        t = np.arange(0, self.tmax, self.dt)
        R1v = self._relax_portal()
        signal = sig.signal_ss(self.S0v, R1v, self.TR, self.FA)
        return utils.sample(xdata, t, signal, self.TS)

    def _conc_liver(self, sum=True):
        pars = self._par_values(kin=True)
        return liver.conc_liver(
            self.ca, dt=self.dt, sum=sum, cv=self.cv, **pars)
    
    def _relax_liver(self):
        t = np.arange(0, self.tmax, self.dt)
        Cl = self._conc_liver(sum=False)
        rp = lib.relaxivity(self.field_strength, 'plasma', self.agent)
        rh = lib.relaxivity(self.field_strength, 'hepatocytes', self.agent)
        if 'IC' in self.kinetics:
          return t, self.R10l + rp*Cl[0, :] + rh*Cl[1, :]
        else:
          return t, self.R10l + rp*Cl

    def _predict_liver(self, xdata):
        t, R1l = self._relax_liver()
        signal = sig.signal_ss(self.S0l, R1l, self.TR, self.FA)
        return utils.sample(xdata, t, signal, self.TS)

    def conc(self, sum=True):
        """Concentrations in aorta. portal vein and liver.

        Args:
            sum (bool, optional): If set to true, the liver concentrations are 
              the sum over both compartments. If set to false, the 
              compartmental concentrations are returned individually. 
              Defaults to True.

        Returns:
            tuple: time points, aorta blood concentrations, portal-venous 
            blood concentrations, liver concentrations.
        """
        t, cb = self._conc_aorta()
        cv = self._conc_portal()
        C = self._conc_liver(sum=sum)
        return t, cb, cv, C

    def relax(self):
        """Relaxation rates in aorta, portal vein and liver.

        Returns:
            tuple: time points, aorta blood R1, portal-venous blood R1, liver 
            blood R1.
        """
        t, R1b = self._relax_aorta()
        R1v = self._relax_portal()
        t, R1l = self._relax_liver()
        return t, R1b, R1v, R1l

    def predict(self, time: tuple) -> tuple:
        """Predict the signals at given time

        Args:
            xdata (tuple): tuple of 3 arrays with time points for aorta, 
              portal vein and liver, in that order. The 3 arrays can be 
              different in length and value.

        Returns:
            tuple: tuple of 3 arrays with signals for aorta, portal vein and 
              liver, in that order. The arrays can be different in length and 
              value but each has to have the same length as its corresponding 
              array of time points.
        """
        # Public interface
        if self._predict is None:
            signala = self._predict_aorta(time[0])
            signalv = self._predict_portal(time[1])
            signall = self._predict_liver(time[2])
            return signala, signalv, signall
        
        # Private interface with different input & output types
        elif self._predict == 'aorta':
            return self._predict_aorta(time)
        elif self._predict == 'portal':
            return self._predict_portal(time)
        elif self._predict == 'liver':
            return self._predict_liver(time)

    def train(self, time: tuple, signal: tuple, **kwargs):
        """Train the free parameters

        Args:
            time (tuple): tuple of 3 arrays with time points for aorta, 
              portal vein and liver, in that order. The arrays can be 
              different in length and value.
            signal (array-like): tuple of 3 arrays with signals for aorta, 
              portal vein and liver, in that order. The arrays can be 
              different in length and value but each has to have the same 
              length as its corresponding array of time points.
            kwargs: any keyword parameters accepted by 
              `scipy.optimize.curve_fit`.

        Returns:
            AortaPortalLiver: the trained model
        """
        # Estimate BAT and S0a from data
        pars = self._par_values(seq=self.sequence)
        if self.sequence == 'SSI':
            Srefa = sig.signal_spgr(1, self.R10a, self.TF, self.TR, self.FA, n0=1)
        else:
            Srefa = sig.signal_ss(1, self.R10a, self.TR, self.FA)
        Srefv = sig.signal_ss(1, self.R10a, self.TR, self.FA)
        Srefl = sig.signal_ss(1, self.R10l, self.TR, self.FA)

        n0 = max([np.sum(time[0] < self.t0), 1])
        self.S0a = np.mean(signal[0][:n0]) / Srefa
        self.S0v = np.mean(signal[1][:n0]) / Srefv
        self.S0l = np.mean(signal[2][:n0]) / Srefl
        self.BAT = time[0][np.argmax(signal[0])] - (1-self.Dhl)*self.Thl
        self.BAT = max([self.BAT, 0])

        # Copy the original free parameters to restore later
        free = deepcopy(self.free)

        # Train free aorta parameters on aorta data
        self._predict = 'aorta'
        pars = list(PARAMS_WHOLE_BODY.keys())
        self.free = {s: free[s] for s in pars if s in free}
        ui.train(self, time[0], signal[0], **kwargs)

        # Train free aorta parameters on portal venous data
        self._predict = 'portal'
        pars = list(PARAMS_PORTAL.keys())
        self.free = {s: free[s] for s in pars if s in free}
        ui.train(self, time[1], signal[1], **kwargs)

        # Train free liver parameters on liver data
        self._predict = 'liver'
        pars = list(PARAMS_LIVER.keys())
        self.free = {s: free[s] for s in pars if s in free}
        ui.train(self, time[2], signal[2], **kwargs)

        # Train all parameters on all data
        self._predict = None
        self.free = free
        return ui.train(self, time, signal, **kwargs)
    

    def plot(self,
             time: tuple,
             signal: tuple,
             xlim=None, ref=None,
             fname=None, show=True):
        """Plot the model fit against data

        Args:
            time (tuple): tuple of 3 arrays with time points for aorta, 
              portal vein and liver, in that order. The two arrays can be 
              different in length and value.
            signal (array-like): tuple of 3 arrays with signals for aorta, 
              portal vein and liver, in that order. The arrays can be 
              different in length and value but each has to have the same 
              length as its corresponding array of time points.
            xlim (array_like, optional): 2-element array with lower and upper 
              boundaries of the x-axis. Defaults to None.
            ref (tuple, optional): Tuple of optional test data in the form 
              (x,y), where x is an array with x-values and y is an array with 
              y-values. Defaults to None.
            fname (path, optional): Filepath to save the image. If no value 
              is provided, the image is not saved. Defaults to None.
            show (bool, optional): If True, the plot is shown. Defaults to 
              True.
        """
        t, cb, cv, C = self.conc(sum=False)
        sig = self.predict((t, t, t))
        fig, ((ax1, ax2), (ax3, ax4), (ax5, ax6)) = plt.subplots(3, 2, figsize=(10, 8))
        fig.subplots_adjust(wspace=0.3)
        _plot_data(t, sig[0], time[0], signal[0],
                        ax1, xlim,
                        color=['lightcoral', 'darkred'],
                        test=None if ref is None else ref[0])
        _plot_data(t, sig[1], time[1], signal[1],
                        ax3, xlim,
                        color=['orchid', 'purple'],
                        test=None if ref is None else ref[1])
        _plot_data(t, sig[2], time[2], signal[2],
                        ax5, xlim,
                        color=['cornflowerblue', 'darkblue'],
                        test=None if ref is None else ref[2],
                        xlabel='Time (min)')
        _plot_conc_aorta(t, cb, ax2, xlim)
        _plot_conc_portal(t, cv, ax4, xlim)
        _plot_conc_liver(self, t, C, ax6, xlim)
        if fname is not None:
            plt.savefig(fname=fname)
        if show:
            plt.show()
        else:
            plt.close()

    def cost(self, time, signal, metric='NRMS') -> float:
        """Return the goodness-of-fit

        Args:
            time (tuple): tuple of 3 arrays with time points for aorta, 
              portal vein and liver, in that order. The arrays can be 
              different in length and value.
            signal (array-like): tuple of 3 arrays with signals for aorta, 
              portal vein and liver, in that order. The arrays can be 
              different in length and value but each has to have the same 
              length as its corresponding array of time points.
            metric (str, optional): Which metric to use (see notes for 
              possible values). Defaults to 'NRMS'.

        Returns:
            float: goodness of fit.

        Notes:

            Available options are: 
            
            - 'RMS': Root-mean-square.
            - 'NRMS': Normalized root-mean-square. 
            - 'AIC': Akaike information criterion. 
            - 'cAIC': Corrected Akaike information criterion for small models.
            - 'BIC': Bayesian information criterion.
        """
        return super().cost(time, signal, metric)




# Helper functions for plotting

def _plot_conc_aorta(t, cb, ax, xlim=None):
    if xlim is None:
        xlim = [t[0], t[-1]]
    ax.set(ylabel='Concentration (mM)',
           xlim=np.array(xlim)/60)
    ax.plot(t/60, 0*t, color='gray')
    ax.plot(t/60, 1000*cb, linestyle='-',
            color='darkred', linewidth=2.0, label='Aorta')
    ax.legend()


def _plot_conc_portal(t, cv, ax, xlim=None):
    if xlim is None:
        xlim = [t[0], t[-1]]
    ax.set(ylabel='Concentration (mM)',
           xlim=np.array(xlim)/60)
    ax.plot(t/60, 0*t, color='gray')
    ax.plot(t/60, 1000*cv, linestyle='-',
            color='purple', linewidth=2.0, label='Portal vein')
    ax.legend()


def _plot_conc_liver(self, t, C, ax, xlim=None):
    color = 'darkblue'
    if xlim is None:
        xlim = [t[0], t[-1]]
    ax.set(xlabel='Time (min)', ylabel='Tissue concentration (mM)',
           xlim=np.array(xlim)/60)
    ax.plot(t/60, 0*t, color='gray')
    if 'IC' in self.kinetics:
        ax.plot(t/60, 1000*C[0, :], linestyle='-.',
                color=color, linewidth=2.0, label='Extracellular')
        ax.plot(t/60, 1000*C[1, :], linestyle='--',
                color=color, linewidth=2.0, label='Hepatocytes')
        ax.plot(t/60, 1000*(C[0, :]+C[1, :]), linestyle='-',
                color=color, linewidth=2.0, label='Liver')
    else:
        ax.plot(t/60, 1000*C, linestyle='-',
                color=color, linewidth=2.0, label='Liver')
    ax.legend()


def _plot_data(t: np.ndarray, sig: np.ndarray,
                    xdata: np.ndarray, ydata: np.ndarray,
                    ax, xlim, color=['black', 'black'],
                    test=None, xlabel=None):
    if xlim is None:
        xlim = [t[0], t[-1]]
    ax.set(xlabel=xlabel, ylabel='MR Signal (a.u.)', 
           xlim=np.array(xlim)/60)
    ax.plot(xdata/60, ydata, marker='o',
            color=color[0], label='fitted data', linestyle='None')
    ax.plot(t/60, sig, linestyle='-',
            color=color[1], linewidth=3.0, label='fit')
    if test is not None:
        ax.plot(np.array(test[0])/60, test[1], color='black',
                marker='D', linestyle='None', label='Test data')
    ax.legend()









PARAMS = {

    # Prediction and training
    't0': {
        'init': 0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Baseline duration',
        'unit': 'sec',
    },
    'dt': {
        'init': 0.5,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Forward model time step',
        'unit': 'sec',
    },
    'tmax': {
        'init': 120,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Maximum acquisition time',
        'unit': 'sec',
    },
    'dose_tolerance': {
        'init': 0.1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Dose tolerance',
        'unit': '',
    },


    # Injection
    'field_strength': {
        'init': 3.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Magnetic field strength',
        'unit': 'T',
    },
    'weight': {
        'init': 70.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Subject weight',
        'unit': 'kg',
    },
    'agent': {
        'init': 'gadoxetate',
        'default_free': False,
        'bounds': None,
        'name': 'Contrast agent',
        'unit': None,
    },
    'dose': {
        'init': lib.ca_std_dose('gadoxetate')/2,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Contrast agent dose',
        'unit': 'mL/kg',
    },
    'rate': {
        'init': 1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Contrast agent injection rate',
        'unit': 'mL/sec',
    },


    # Signal
    'R10a': {
        'init': 1/lib.T1(3.0, 'blood'),
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Aorta baseline R1',
        'unit': 'Hz',
    },
    'S0a': {
        'init': 1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Aorta signal scale factor',
        'unit': 'a.u.',
    },
    'R10l': {
        'init': 1/lib.T1(3.0, 'liver'),
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Liver baseline R1',
        'unit': 'Hz',
    },
    'S0l': {
        'init': 1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Liver signal scale factor',
        'unit': 'a.u.',
    },
    'S0v': {
        'init': 1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Portal venous signal scale factor',
        'unit': 'a.u.',
    },

}


PARAMS_WHOLE_BODY = {

    # Body
    'BAT': {
        'init': 60,
        'default_free': True,
        'bounds': [0, np.inf],
        'name': 'Bolus arrival time',
        'unit': 'sec',
    },
    'CO': {
        'init': 100,
        'default_free': True,
        'bounds': [0, 300],
        'name': 'Cardiac output',
        'unit': 'mL/sec',
    },
    'Thl': {
        'init': 10,
        'default_free': True,
        'bounds': [0, 30],
        'name': 'Heart-lung mean transit time',
        'unit': 'sec',
    },
    'Dhl': {
        'init': 0.2,
        'default_free': True,
        'bounds': [0.05, 0.95],
        'name': 'Heart-lung dispersion',
        'unit': '',
    },
    'To': {
        'init': 20,
        'default_free': True,
        'bounds': [0, 60],
        'name': 'Organs blood mean transit time',
        'unit': 'sec',
    },
    'Eo': {
        'init': 0.15,
        'default_free': True,
        'bounds': [0, 0.5],
        'name': 'Organs extraction fraction',
        'unit': '',
    },
    'Toe': {
        'init': 120,
        'default_free': True,
        'bounds': [0, 800],
        'name': 'Organs extravascular mean transit time',
        'unit': 'sec',
    },
    'Eb': {
        'init': 0.05,
        'default_free': True,
        'bounds': [0.01, 0.15],
        'name': 'Body extraction fraction',
        'unit': '',
    },
}


PARAMS_SEQUENCE = {
    'TR': {
        'init': 0.005,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Repetition time',
        'unit': 'sec',
    },
    'FA': {
        'init': 15.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Flip angle',
        'unit': 'deg',
    },
    'TF': {
        'init': 0.5,
        'default_free': True,
        'bounds': [0, 5],
        'name': 'Aorta inflow time',
        'unit': 'sec',
    },
    'TS': {
        'init': None,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Sampling time',
        'unit': 'sec',
    },
}

PARAMS_PORTAL = {
    'Tg': {
        'init': 15.0,
        'default_free': True,
        'bounds': [0.1, 60],
        'name': 'Gut mean transit time',
        'unit': 'sec',
    },
    'Dg': {
        'init': 0.85,
        'default_free': True,
        'bounds': [0, 1],
        'name': 'Gut dispersion',
        'unit': '',
    },
    'uv': {
        'init': 1.0,
        'default_free': True,
        'bounds': [0, 1],
        'name': 'Portal vein volume fraction',
        'unit': '',
    },

}

PARAMS_LIVER = {
    'H': {
        'init': 0.45,
        'default_free': False,
        'bounds': [0, 1],
        'name': 'Hematocrit',
        'unit': '',
    },
    've': {
        'init': 0.3,
        'default_free': True,
        'bounds': [0.01, 0.6],
        'name': 'Liver extracellular volume fraction',
        'unit': 'mL/cm3',
    },
    'Fp': {
        'init': 0.01,
        'default_free': True,
        'bounds': [0.00, 0.1],
        'name': 'Liver plasma flow',
        'unit': 'mL/sec/cm3',
    },
    'fa': {
        'init': 0.2,
        'default_free': True,
        'bounds': [0.0, 1.0],
        'name': 'Arterial flow fraction',
        'unit': '',
    },
    'Ta': {
        'init': 0.5,
        'default_free': True,
        'bounds': [0.0, 3],
        'name': 'Arterial transit time',
        'unit': 'sec',
    },
    'khe': {
        'init': 0.003,
        'default_free': True,
        'bounds': [0.0, 0.1],
        'name': 'Hepatocellular uptake rate',
        'unit': 'mL/sec/cm3',
    },
    'khe_i': {
        'init': 0.003,
        'default_free': True,
        'bounds': [0.0, 0.1],
        'name': 'Initial hepatocellular uptake rate',
        'unit': 'mL/sec/cm3',
        'pixel_par': True,
    },
    'khe_f': {
        'init': 0.003,
        'default_free': True,
        'bounds': [0.0, 0.1],
        'name': 'Final hepatocellular uptake rate',
        'unit': 'mL/sec/cm3',
        'pixel_par': True,
    },
    'Th': {
        'init': 30*60,
        'default_free': True,
        'bounds': [10*60, 10*60*60],
        'name': 'Hepatocellular mean transit time',
        'unit': 'sec',
    },
    'Th_i': {
        'init': 30*60,
        'default_free': True,
        'bounds': [10*60, 10*60*60],
        'name': 'Initial hepatocellular mean transit time',
        'unit': 'sec',
        'pixel_par': True,
    },
    'Th_f': {
        'init': 30*60,
        'default_free': True,
        'bounds': [10*60, 10*60*60],
        'name': 'Final hepatocellular mean transit time',
        'unit': 'sec',
        'pixel_par': True,
    },
    'vol': {
        'init': None,
        'default_free': False,
        'bounds': [0, 10000],
        'name': 'Liver volume',
        'unit': 'cm3',
    },
}

PARAMS_DERIVED = {
    'Fv': {
        'name': 'Portal venous blood flow',
        'unit': 'mL/sec/cm3',
    },
    'Fa': {
        'name': 'Arterial venous blood flow',
        'unit': 'mL/sec/cm3',
    },
    'Te': {
        'name': 'Extracellular mean transit time',
        'unit': 'sec',
    },
    'kbh': {
        'name': 'Biliary excretion rate',
        'unit': 'mL/sec/cm3',
    },
    'kbh_i': {
        'name': 'Initial biliary excretion rate',
        'unit': 'mL/sec/cm3',
    },
    'kbh_f': {
        'name': 'Final biliary excretion rate',
        'unit': 'mL/sec/cm3',
    },
    'Kbh': {
        'name': 'Biliary tissue excretion rate',
        'unit': 'mL/sec/cm3',
    },
    'Khe': {
        'name': 'Hepatocellular tissue uptake rate',
        'unit': 'mL/sec/cm3',
    },
    'E': {
        'name': 'Liver extraction fraction',
        'unit': '',
    },
    'Ktrans': {
        'name': 'Hepatic plasma clearance',
        'unit': 'mL/sec/cm3',
    },
    'CL': {
        'name': 'Liver blood clearance',
        'unit': 'mL/sec',
    }, 
}


def _div(a, b):
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.divide(a, b)


import os
import shutil
import numpy as np
import dcmri as dc


# Debugging mode
VERBOSE = 1
SHOW = True

# Production mode
VERBOSE = 0
SHOW = False



def test_ui_tissue_ls():

    # Create data
    time, aif, roi, gt = dc.fake_tissue()
    gt['ve'] = gt['vp'] + gt['vi'] if gt['PS'] > 0 else gt['vp']

    params = {
        'dt': time[1],
        'sequence': 'SS',
        'r1': dc.relaxivity(3,'blood','gadodiamide'),
        'TR': 0.005,
        'FA': 15,
        'R10a': 1/dc.T1(3.0,'blood'),
        'R10': 1/dc.T1(3.0,'muscle'),
    }

    # Train model and check results
    model = dc.TissueLS(**params)
    model.train(roi, aif, n0=10, tol=0.01)
    model.plot(roi, show=SHOW)
    pars = model.params()
    
    assert np.abs(pars['Fb']-gt['Fb'])/gt['Fb'] < 0.1
    assert np.abs(pars['ve']-gt['ve'])/gt['ve'] < 0.2



def test_ui_tissue_ls_array():

    # Create ground truth data
    n=16
    time, signal, aif, gt = dc.fake_brain(n=n, verbose=VERBOSE)
    with np.errstate(divide='ignore'):
        R10 = np.where(gt['T1']==0, 0, 1/gt['T1'])
    vmin = {'Fb':0, 've':0, 'S0':0}
    vmax = {'Fb':0.02, 've':0.2, 'S0':np.amax(gt['S0'])}
    gt['ve'] = np.where(gt['PS'] > 0, gt['vp'] + gt['vi'], gt['vp'])
    gt['Te'] = np.where(gt['Fp']==0, 0, gt['ve']/gt['Fp'])

    # Define model parameters
    params = {
        'dt': time[1],
        'sequence': 'SS',
        'r1': dc.relaxivity(3,'blood','gadodiamide'),
        'TR': 0.005,
        'FA': 15,
        'R10a': 1/dc.T1(3.0,'blood'),
        'R10': R10,
    }

    # Train array and show results
    image = dc.TissueLSArray((n,n), **params)
    image.train(signal, aif, n0=10, tol=0.01)
    image.plot(vmin=vmin, vmax=vmax, ref=gt, show=SHOW)
    assert np.linalg.norm(image.params('Fb')-gt['Fb'])/np.linalg.norm(gt['Fb']) < 0.1

    # Repeat with linear model (different than ground truth model -> less accurate result)
    params['sequence'] = 'lin'
    image = dc.TissueLSArray((n,n), **params)
    image.train(signal, aif, n0=10, tol=0.01)
    image.plot(vmin=vmin, vmax=vmax, ref=gt, show=SHOW)
    assert np.linalg.norm(image.params('Fb')-gt['Fb'])/np.linalg.norm(gt['Fb']) < 1


if __name__ == "__main__":

    # test_ui_tissue_ls()
    test_ui_tissue_ls_array()

    print('All ui_tissue tests passed!!')



import copy
import numpy as np

AGENTS = [
    'gadoxetate',
    'gadobutrol',
    'gadopentetate',
    'gadobenate',
    'gadodiamide',
    'gadoterate',
    'gadoteridol',
    'gadopiclenol',
]


def ca_injection(t: np.ndarray, weight: float, conc: float,
                dose: float, rate: float, t0: float) -> np.ndarray:
    """Contrast agent flux (mmol/sec) generated by step injection.

    Args:
        t (numpy.ndarray): time points in sec where the flux is to be calculated.
        weight (float): weight of the subject in kg.
        conc (float): concentration of the contrast agent in mmol/mL.
        dose (float): injected dose in mL/kg body weight.
        rate (float): rate of contrast agent injection in mL/sec.
        t0 (float): start of the injection in sec.

    Raises:
        ValueError: if the injection duration is zero, or smaller than the time step of the time array.

    Returns:
        numpy.ndarray: contrast agent flux for each time point in units of mmol/sec.

    Example:

        >>> import numpy as np
        >>> import dcmri as dc

        Create an array of time points covering 20sec in steps of 1.5sec.

        >>> t = np.arange(0, 20, 1.5)

        Inject a dose of 0.2 mL/kg bodyweight at a rate of 3mL/sec starting at t=5sec.
        For a subject weighing 70 kg and a contrast agent with concentration 0.5M, this produces the flux:

        >>> dc.ca_injection(t, 70, 0.5, 5, 0.2, 3)
        array([0. , 0. , 0. , 0. , 1.5, 1.5, 1.5, 0. , 0. , 0. , 0. , 0. , 0. ,0. ])
    """

    # Get timings
    duration = weight * dose / rate     # sec = kg * (mL/kg) / (mL/sec)
    dt = np.amin(t[1:] - t[:-1])

    # Check consistency of timings
    if dose > 0:
        if duration == 0:
            msg = 'Invalid input variables. '
            msg += 'The injection duration is zero.'
            raise ValueError(msg)
        if dt >= duration:
            msg = 'Invalid input variables. \n'
            msg += 'The smallest time step dt (' + str(
                dt) + ' sec) is larger than the injection duration 1 (' + str(duration) + 'sec). \n'
            msg += 'We would recommend dt to be at least 5 times smaller.'
            raise ValueError(msg)

    # Build flux
    Jmax = conc * rate                # mmol/sec = (mmol/ml) * (ml/sec)
    J = np.zeros(t.size)
    J[(0 < t) & (t < duration)] = Jmax
    return np.interp(t - t0, t, J, left=0)


def ca_conc(agent: str) -> float:
    """Contrast agent concentration

    Args:
        agent (str): Generic contrast agent name, all lower case. Examples are 'gadobutrol', 'gadobenate', etc.

    Raises:
        ValueError: If no data are available for the agent.

    Returns:
        float: concentration in mmol/mL

    Sources:
        - `mri questions <https://mriquestions.com/so-many-gd-agents.html>`_
        - `gadoxetate <https://www.bayer.com/sites/default/files/2020-11/primovist-pm-en.pdf>`_
        - `gadobutrol <https://www.medicines.org.uk/emc/product/2876/smpc#gref>`_

    Example:

        Print the concentration of the agents gadobutrol and gadoterate:

    .. exec_code::

        import dcmri as dc

        print('gadobutrol is available in a solution of', dc.ca_conc('gadobutrol'), 'M')
        print('gadoterate is available in a solution of', dc.ca_conc('gadoterate'), 'M')
    """
    if agent == 'gadoxetate':
        return 0.25     # mmol/mL
    if agent == 'gadobutrol':
        return 1.0      # mmol/mL
    if agent in [
        'gadopentetate',
        'gadobenate',
        'gadodiamide',
        'gadoterate',
        'gadoteridol',
        'gadopiclenol',
    ]:
        return 0.5  # mmol/mL
    raise ValueError(
        f"No concentration data for contrast agent {agent}."
        f"Possible values are {AGENTS}."
    )


def ca_std_dose(agent: str) -> float:
    """Standard injection volume (dose) in mL per kg body weight.

    Args:
        agent (str): Generic contrast agent name, all lower case. Examples are 'gadobutrol', 'gadobenate', etc.

    Raises:
        ValueError: If no data are available for the agent.

    Returns:
        float: Standard injection volume in mL/kg.

    Note:

        Available agents:
            - gadoxetate
            - gadobutrol
            - gadopiclenol
            - gadopentetate
            - gadobenate
            - gadodiamide
            - gadoterate
            - gadoteridol

        Sources:
            - `mri questions <https://mriquestions.com/so-many-gd-agents.html>`_
            - `gadoxetate <https://www.bayer.com/sites/default/files/2020-11/primovist-pm-en.pdf>`_
            - `gadobutrol <https://www.medicines.org.uk/emc/product/2876/smpc#gref>`_

    Example:

        >>> import dcmri as dc
        >>> print('The standard clinical dose of gadobutrol is', dc.ca_std_dose('gadobutrol'), 'mL/kg')
        The standard clinical dose of gadobutrol is 0.1 mL/kg
    """
    # """Standard dose in mL/kg""" # better in mmol/kg, or offer it as an option
    if agent == 'gadoxetate':
        # https://www.bayer.com/sites/default/files/2020-11/primovist-pm-en.pdf
        return 0.1  # mL/kg
    if agent == 'gadobutrol':
        return 0.1      # mL/kg
    if agent == 'gadopiclenol':
        return 0.1      # mL/kg
    if agent in [
            'gadopentetate',
            'gadobenate',
            'gadodiamide',
            'gadoterate',
            'gadoteridol',
    ]:
        return 0.2      # mL/kg  # 0.5 mmol/mL = 0.1 mmol/kg
    raise ValueError(
        f"No data available for contrast agent {agent}."
        f"Possible values are {AGENTS}."
    )


def relaxivity(field_strength=3.0, tissue='plasma',
               agent='gadoxetate', type='T1') -> float:
    """Contrast agent relaxivity values in units of Hz/M

    Args:
        field_strength (float, optional): Field strength in Tesla. Defaults to 3.0.
        tissue (str, optional): Tissue type - options are 'plasma', 'hepatocytes'. Defaults to 'plasma'.
        agent (str, optional): Generic contrast agent name, all lower case. Examples are 'gadobutrol', 'gadobenate', etc.. Defaults to 'gadoxetate'.
        type (str, optional): transverse (T2 or T2*) or longitudinal (T1) relaxivity. Defaults to 'T1'.

    Returns:
        float: relaxivity in Hz/M or 1/(sec*M)

    Note:

        This library is in construction and not all known values are currently available through this function.

        Available contrasts:
            - T1

        Available tissues:
            - blood
            - plasma
            - hepatocytes

        Available agents:
            - gadopentetate
            - gadobutrol
            - gadoteridol
            - gadobenade
            - gadoterate
            - gadodiamide
            - mangafodipir
            - gadoversetamide
            - ferucarbotran
            - ferumoxide
            - gadoxetate

        Available field strengths:
            - 0.47
            - 1.5
            - 3
            - 4.7
            - 7.0 (gadoxetate)
            - 9.0 (gadoxetate)

        Sources:
            - Rohrer M, et al. Comparison of Magnetic Properties of MRI Contrast Media Solutions at Different Magnetic Field Strengths. Investigative Radiology 40(11):p 715-724, November 2005. DOI: `10.1097/01.rli.0000184756.66360.d3 <https://journals.lww.com/investigativeradiology/FullText/2005/11000/Comparison_of_Magnetic_Properties_of_MRI_Contrast.5.aspx>`_
            - Szomolanyi P, et al. Comparison of the Relaxivities of Macrocyclic Gadolinium-Based Contrast Agents in Human Plasma at 1.5, 3, and 7 T, and Blood at 3 T. Invest Radiol. 2019 Sep;54(9):559-564. doi: `10.1097/RLI.0000000000000577 <https://pubmed.ncbi.nlm.nih.gov/31124800/>`_

    Example:

        >>> import dcmri as dc
        >>> print('The plasma relaxivity of gadobutrol at 3T is', 1e-3*dc.relaxivity(3.0, 'plasma', 'gadobutrol'), 'Hz/mM')
        The plasma relaxivity of gadobutrol at 3T is 5.0 Hz/mM
    """
    # Blood and plasma have (theoretically) the same relaxivity
    if tissue == 'blood':
        tissue = 'plasma'
    rel = {}
    rel['T1'] = {
        'plasma': {
            'gadopentetate': {  # Magnevist
                0.47: 3.8,
                1.5: 4.1,
                3.0: 3.7,
                4.7: 3.8,
            },
            'gadobutrol': {  # Gadovist
                0.47: 6.1,
                1.5: 5.2,
                3.0: 5.0,
                4.7: 4.7,
            },
            'gadoteridol': {  # Prohance
                0.47: 4.8,
                1.5: 4.1,
                3.0: 3.7,
                4.7: 3.7,
            },
            'gadobenade': {  # Multihance
                0.47: 9.2,
                1.5: 6.3,
                3.0: 5.5,
                4.7: 5.2,
            },
            'gadoterate': {  # Dotarem
                0.47: 4.3,
                1.5: 3.6,
                3.0: 3.5,
                4.7: 3.3,
            },
            'gadodiamide': {  # Omniscan
                0.47: 4.4,
                1.0: 4.35,  # Interpolated
                1.5: 4.3,
                3.0: 4.0,
                4.7: 3.9,
            },
            'mangafodipir': {  # Teslascan
                0.47: 3.6,
                1.5: 3.6,
                3.0: 2.7,
                4.7: 2.2,
            },
            'gadoversetamide': {  # Optimark
                0.47: 5.7,
                1.5: 4.7,
                3.0: 4.5,
                4.7: 4.4,
            },
            'ferucarbotran': {  # Resovist
                0.47: 15,
                1.5: 7.4,
                3.0: 3.3,
                4.7: 1.7,
            },
            'ferumoxide': {  # Feridex
                1.5: 4.5,
                3.0: 2.7,
                4.7: 1.2,
            },
            'gadoxetate': {  # Primovist
                0.47: 8.7,
                1.5: 8.1,
                3.0: 6.4,
                4.7: 6.4,
                7.0: 6.2,
                9.0: 6.1
            },
        },
    }
    rel['T1']['hepatocytes'] = copy.deepcopy(rel['T1']['plasma'])
    rel['T1']['hepatocytes']['gadoxetate'] = {
        1.5: 14.6,
        3.0: 9.8,
        4.7: 7.6,
        7.0: 6.0,
        9.0: 6.1,
    }
    try:
        return 1000 * rel[type][tissue][agent][field_strength]
    except KeyError:
        msg = 'No relaxivity data for ' + agent + \
            ' at ' + str(field_strength) + ' T.'
        raise ValueError(msg)


def T1(field_strength=3.0, tissue='blood', Hct=0.45) -> float:
    """T1 value of selected tissue types.

    Values are taken from literature, mostly from `Stanisz et al 2005 <https://doi.org/10.1002/mrm.20605>`_

    Args:
        field_strength (float, optional): Field strength in Tesla (see below for options). Defaults to 3.0.
        tissue (str, optional): Tissue type (see below for options). Defaults to 'blood'.
        Hct (float, optional): Hematocrit value - ignored when tissue is not blood. Defaults to 0.45.

    Raises:
        ValueError: If the requested T1 values are not available.

    Returns:
        float: T1 values in sec

    Note:

        This library is in construction and not all known values are currently available through this function.

        Available tissue types:
            - skin
            - bone marrow
            - csf
            - muscle
            - heart
            - cartilage
            - white matter
            - gray matter
            - optic nerve
            - spinal cord
            - blood
            - spleen
            - liver
            - kidney

        Available field strengths:
            - 1.5
            - 3
            - 4.7 (spleen, liver)
            - 7.0 (spleen, liver)
            - 9.0 (spleen, liver)

    Example:

        >>> import dcmri as dc
        >>> print('The T1 of liver at 1.5T is', 1e3*dc.T1(1.5, 'liver'), 'msec')
        The T1 of liver at 1.5T is 602.0 msec
    """
    # H. M. Gach, C. Tanase and F. Boada, "2D & 3D Shepp-Logan Phantom
    # Standards for MRI," 2008 19th International Conference on Systems
    # Engineering, Las Vegas, NV, USA, 2008, pp. 521-526, doi:
    # 10.1109/ICSEng.2008.15.

    T1val = {
        'skin': {  # Gach 2008 (scalp)
            1.5: 0.324 * (1.5**0.137),
            3.0: 0.324 * (3.0**0.137),
        },
        'bone marrow': {  # Gach 2008
            1.5: 0.533 * (1.5**0.088),
            3.0: 0.533 * (3.0**0.088),
        },
        'csf': {  # Gach 2008
            1.5: 4.20,
            3.0: 4.20,
        },
        'muscle': {
            1.5: 1.008,
            3.0: 1.412,
        },
        'heart': {
            1.5: 1.030,
            3.0: 1.471,
        },
        'cartilage': {
            1.5: 1.024,
            3.0: 1.168,
        },
        'white matter': {
            1.5: 0.884,
            3.0: 1.084,
        },
        'gray matter': {
            1.5: 1.124,
            3.0: 1.820,
        },
        'optic nerve': {
            1.5: 0.815,
            3.0: 1.083,
        },
        'spinal cord': {
            1.5: 0.745,
            3.0: 0.993,
        },
        'blood': {
            1.0: 1.378,  # Extrapolated
            1.5: 1.441,
            3.0: 1 / (0.52 * Hct + 0.38),  # Lu MRM 2004
            4.7: 1 / 1.70,  # https://cds.ismrm.org/ismrm-2002/PDF4/1048.PDF
            7.0: 1 / 2.29,  # 10.1016/j.mri.2012.08.008
        },
        'spleen': {
            4.7: 1 / 0.631,
            7.0: 1 / 0.611,
            9.0: 1 / 0.600,
        },
        'liver': {
            1.5: 0.602,  # liver R1 in 1/sec (Waterton 2021)
            3.0: 0.752,  # liver R1 in 1/sec (Waterton 2021)
            # liver R1 in 1/sec (Changed from 1.285 on 06/08/2020)
            4.7: 1 / 1.281,
            # liver R1 in 1/sec (Changed from 0.8350 on 06/08/2020)
            7.0: 1 / 1.109,
            # per sec - liver R1 (https://doi.org/10.1007/s10334-021-00928-x)
            9.0: 1 / 0.920,
        },
        'kidney': {
            # Reference values average over cortex and medulla from Cox et al
            # https://academic.oup.com/ndt/article/33/suppl_2/ii41/5078406
            1.0: 1.017,  # Extrapolated
            1.5: (1.024 + 1.272) / 2,
            3.0: (1.399 + 1.685) / 2,
        },
    }
    try:
        return T1val[tissue][field_strength]
    except BaseException:
        msg = 'No T1 values for ' + tissue + \
            ' at ' + str(field_strength) + ' T.'
        raise ValueError(msg)


def T2(field_strength=3.0, tissue='gray matter') -> float:
    """T2 value of selected tissue types.

    Values are taken from `Gach et al 2008 <https://ieeexplore.ieee.org/document/4616690>`_

    Args:
        field_strength (float, optional): Field strength in Tesla. Defaults to 3.0.
        tissue (str, optional): Tissue type. Defaults to 'gray matter'.

    Raises:
        ValueError: If the requested T2 values are not available.

    Returns:
        float: T2 values in sec

    Note:

        This library is in construction and not all known values are currently available through this function.

        Available tissue types:
            - skin
            - bone marrow
            - csf
            - white matter
            - gray matter

        Available field strengths:
            - 1.5
            - 3

    Example:

        >>> import dcmri as dc
        >>> print('The T2 of skin at 1.5T is', 1e3*dc.T2(1.5, 'skin'), 'msec')
        The T2 of skin at 1.5T is 70.0 msec
    """
    # H. M. Gach, C. Tanase and F. Boada, "2D & 3D Shepp-Logan Phantom
    # Standards for MRI," 2008 19th International Conference on Systems
    # Engineering, Las Vegas, NV, USA, 2008, pp. 521-526, doi:
    # 10.1109/ICSEng.2008.15.

    T2val = {
        'skin': {  # Gach 2008 (scalp)
            1.5: 0.07,
            3.0: 0.07,
        },
        'bone marrow': {  # Gach 2008
            1.5: 0.05,
            3.0: 0.05,
        },
        'csf': {  # Gach 2008
            1.5: 1.99,
            3.0: 1.99,
        },
        'white matter': {
            1.5: 0.08,
            3.0: 0.08,
        },
        'gray matter': {
            1.5: 0.1,
            3.0: 0.1,
        },
    }
    try:
        return T2val[tissue][field_strength]
    except BaseException:
        msg = 'No T2 values for ' + tissue + \
            ' at ' + str(field_strength) + ' T.'
        raise ValueError(msg)


def PD(tissue='gray matter') -> float:
    """Relative proton density (PD) value of selected tissue types.

    Values are taken from `Gach et al 2008 <https://ieeexplore.ieee.org/document/4616690>`_

    Args:
        tissue (str, optional): Tissue type. Defaults to 'gray matter'.

    Raises:
        ValueError: If the requested PD values are not available.

    Returns:
        float: PD values (dimensionless, range 0-1)

    Note:

        This library is in construction and not all known values are currently available through this function.

        Available tissue types:
            - skin
            - bone marrow
            - csf
            - white matter
            - gray matter

    Example:

        >>> import dcmri as dc
        >>> print('The PD of skin is', dc.PD('skin'))
        The PD of skin is 0.8
    """
    # H. M. Gach, C. Tanase and F. Boada, "2D & 3D Shepp-Logan Phantom
    # Standards for MRI," 2008 19th International Conference on Systems
    # Engineering, Las Vegas, NV, USA, 2008, pp. 521-526, doi:
    # 10.1109/ICSEng.2008.15.

    PDval = {
        'skin': 0.8,
        'bone marrow': 0.12,
        'csf': 0.98,
        'white matter': 0.617,
        'gray matter': 0.745,
    }
    try:
        return PDval[tissue]
    except BaseException:
        msg = 'No PD values for ' + tissue
        raise ValueError(msg)


def perfusion(parameter='Fb', tissue='gray matter') -> float:
    """perfusion parameters of selected tissue types.

    Args:
        parameter (str, optional): perfusion parameter. Options are 'Fb' (blood flow), 'vb' (Blood volume), 'PS' (permeability-surface area product) and 'vi' (interstitial volume). Defaults to 'Fb'.
        tissue (str, optional): Tissue type. Defaults to 'gray matter'.

    Raises:
        ValueError: If the requested parameter values are not available.

    Returns:
        float: parameter value in standard units

    Note:

        This library is in construction and not all known values are currently available through this function.

        Available tissue types:
            - skin
            - bone marrow
            - csf
            - white matter
            - gray matter

    Example:

        >>> import dcmri as dc
        >>> print('The BF of gray matter is', dc.perfusion('Fb', 'gray matter'), 'mL/sec/mL')
        The BF of gray matter is 0.01 mL/sec/mL
    """
    if parameter == 'Fb':
        val = {
            'skin': 0.005,  # 5 kg/s/m**3 = 5000mL/sec/100*100*100 mL = 0.005mL/sec/mL
            'bone marrow': 0.0013,  # 0.08 ml/ml/min
            'csf': 0.0,
            'white matter': 0.0033,
            'gray matter': 0.01,
        }
    elif parameter == 'vb':
        val = {
            'skin': 0.03,
            'bone marrow': 0.25,
            'csf': 0.0,
            'white matter': 0.02,
            'gray matter': 0.05,
        }
    elif parameter == 'PS':  # Needs verification
        val = {
            'skin': 0.001,
            'bone marrow': 0.0002,
            'csf': 0.0,
            'white matter': 0.0,
            'gray matter': 0.0,
        }
    elif parameter == 'vi':  # Needs verification
        val = {
            'skin': 0.03,
            'bone marrow': 0.2,
            'csf': 0.0,
            'white matter': 0.3,
            'gray matter': 0.35,
        }
    try:
        return val[tissue]
    except BaseException:
        msg = 'No ' + parameter + ' values for ' + tissue
        raise ValueError(msg)


def _ellipse(array, center, axes, angle, value):
    n = array.shape[0]
    major = np.amax(axes)
    minor = np.amin(axes)
    c = np.sqrt(major**2 - minor**2)
    dx = np.cos(angle * np.pi / 180)
    dy = np.sin(angle * np.pi / 180)
    f1 = [center[0] + dx * c, center[1] + dy * c]
    f2 = [center[0] - dx * c, center[1] - dy * c]
    d = 2.0 / n
    for i in range(n):
        for j in range(n):
            x = d * (i - (n - 1) / 2)
            y = d * (j - (n - 1) / 2)
            d1 = np.sqrt((x - f1[0])**2 + (y - f1[1])**2)
            d2 = np.sqrt((x - f2[0])**2 + (y - f2[1])**2)
            if d1 + d2 <= 2 * major:
                array[i, j] = value
    return array


def _shepp_logan_mask(n=256) -> dict:

    mask = {}

    # Background
    array = np.ones((n, n)).astype(bool)
    array = _ellipse(array, [0, 0], [0.72, 0.95], 0, 0)
    mask['background'] = np.flip(array, axis=0)

    # 1 Scalp
    array = _ellipse(np.zeros((n, n)).astype(
        bool), [0, 0], [0.72, 0.95], 0, 1)
    array = _ellipse(array, [0, 0], [0.69, 0.92], 0, 0)
    mask['scalp'] = np.flip(array, axis=0)

    # 2 Bone and marrow
    array = _ellipse(np.zeros((n, n)).astype(
        bool), [0, 0], [0.69, 0.92], 0, 1)
    array = _ellipse(array, [-0.0184, 0], [0.6624, 0.874], 0, 0)
    mask['bone'] = np.flip(array, axis=0)

    # 3 CSF
    array = _ellipse(np.zeros((n, n)).astype(bool),
                     [-0.0184, 0], [0.6624, 0.874], 0, 1)
    array = _ellipse(array, [-0.0184, 0], [0.6524, 0.864], 0, 0)
    mask['CSF skull'] = np.flip(array, axis=0)

    # 4 Gray matter
    array = _ellipse(np.zeros((n, n)).astype(bool),
                     [-0.0184, 0], [0.6524, 0.864], 0, 1)
    array = _ellipse(array, [0, -0.22], [0.41, 0.16], 180 - 18, 0)
    array = _ellipse(array, [0, 0.22], [0.31, 0.11], 180 + 18, 0)
    array = _ellipse(array, [0.35, 0], [0.21, 0.25], 0, 0)
    array = _ellipse(array, [0.1, 0], [0.046, 0.046], 0, 0)
    array = _ellipse(array, [-0.605, -0.08], [0.046, 0.023], 90, 0)
    array = _ellipse(array, [-0.605, 0.06], [0.023, 0.046], 0, 0)
    array = _ellipse(array, [-0.1, 0.0], [0.046, 0.046], 0, 0)
    array = _ellipse(array, [-0.605, 0.0], [0.023, 0.023], 0, 0)
    array = _ellipse(array, [-0.83, 0], [0.05, 0.05], 0, 0)
    array = _ellipse(array, [0.66, 0], [0.02, 0.02], 0, 0)
    mask['gray matter'] = np.flip(array, axis=0)

    # 5 CSF
    array = _ellipse(np.zeros((n, n)).astype(bool), [
                     0, -0.22], [0.41, 0.16], 180 - 18, 1)
    array = _ellipse(array, [0.35, 0], [0.21, 0.25], 0, 0)
    array = _ellipse(array, [-0.1, 0.0], [0.046, 0.046], 0, 0)
    mask['CSF right'] = np.flip(array, axis=0)

    # 6 CSF
    array = _ellipse(np.zeros((n, n)).astype(bool), [
                     0, 0.22], [0.31, 0.11], 180 + 18, 1)
    mask['CSF left'] = np.flip(array, axis=0)

    # 7 Tumor
    array = _ellipse(np.zeros((n, n)).astype(
        bool), [0.35, 0], [0.21, 0.25], 0, 1)
    array = _ellipse(array, [0.1, 0], [0.046, 0.046], 0, 0)
    mask['tumor 1'] = np.flip(array, axis=0)

    # 8 Tumor
    array = _ellipse(np.zeros((n, n)).astype(bool),
                     [0.1, 0], [0.046, 0.046], 0, 1)
    mask['tumor 2'] = np.flip(array, axis=0)

    # 9 Tumor
    array = _ellipse(np.zeros((n, n)).astype(bool),
                     [-0.605, -0.08], [0.046, 0.023], 90, 1)
    mask['tumor 4'] = np.flip(array, axis=0)

    # 10 Tumor
    array = _ellipse(np.zeros((n, n)).astype(bool),
                     [-0.605, 0.06], [0.023, 0.046], 0, 1)
    mask['tumor 6'] = np.flip(array, axis=0)

    # 11 Tumor
    array = _ellipse(np.zeros((n, n)).astype(bool),
                     [-0.1, 0.0], [0.046, 0.046], 0, 1)
    mask['tumor 3'] = np.flip(array, axis=0)

    # 12 Tumor
    array = _ellipse(np.zeros((n, n)).astype(bool),
                     [-0.605, 0.0], [0.023, 0.023], 0, 1)
    mask['tumor 5'] = np.flip(array, axis=0)

    # 13 Sagittal sinus
    array = _ellipse(np.zeros((n, n)).astype(bool),
                     [-0.83, 0], [0.05, 0.05], 0, 1)
    mask['sagittal sinus'] = np.flip(array, axis=0)

    # 14 Anterior cerebral artery
    array = _ellipse(np.zeros((n, n)).astype(
        bool), [0.66, 0], [0.02, 0.02], 0, 1)
    mask['anterior artery'] = np.flip(array, axis=0)

    return mask


def _shepp_logan(param, n=256, B0=3) -> np.ndarray:
    if param == 'PD':
        scalp = PD('skin')
        bone = PD('bone marrow')
        csf = PD('csf')
        gm = PD('gray matter')
        tumor = 0.95
        blood = 0.9
    elif param == 'T2':
        scalp = T2(B0, 'skin')
        bone = T2(B0, 'bone marrow')
        csf = T2(B0, 'csf')
        gm = T2(B0, 'gray matter')
        tumor = 0.25
        blood = 0.1
    elif param == 'T1':
        scalp = T1(B0, 'skin')
        bone = T1(B0, 'bone marrow')
        csf = T1(B0, 'csf')
        gm = T1(B0, 'gray matter')
        tumor = 0.926 * (B0**0.217)
        blood = T1(B0, 'blood')
    elif param == 'Fb':
        scalp = perfusion('Fb', 'skin')
        bone = perfusion('Fb', 'bone marrow')
        csf = perfusion('Fb', 'csf')
        gm = perfusion('Fb', 'gray matter')
        tumor = 0.02
        blood = 0
    elif param == 'vb':
        scalp = perfusion('vb', 'skin')
        bone = perfusion('vb', 'bone marrow')
        csf = perfusion('vb', 'csf')
        gm = perfusion('vb', 'gray matter')
        tumor = 0.1
        blood = 1
    elif param == 'PS':
        scalp = perfusion('PS', 'skin')
        bone = perfusion('PS', 'bone marrow')
        csf = perfusion('PS', 'csf')
        gm = perfusion('PS', 'gray matter')
        tumor = 0.001
        blood = 0
    elif param == 'vi':
        scalp = perfusion('vi', 'skin')
        bone = perfusion('vi', 'bone marrow')
        csf = perfusion('vi', 'csf')
        gm = perfusion('vi', 'gray matter')
        tumor = 0.3
        blood = 0

    array = np.zeros((n, n)).astype(np.float64)
    # 1 Scalp
    array = _ellipse(array, [0, 0], [0.72, 0.95], 0, scalp)
    # 2 Bone and marrow
    array = _ellipse(array, [0, 0], [0.69, 0.92], 0, bone)
    # 3 CSF
    array = _ellipse(array, [-0.0184, 0], [0.6624, 0.874], 0, csf)
    # 4 Gray matter
    array = _ellipse(array, [-0.0184, 0], [0.6524, 0.864], 0, gm)
    # 5 CSF
    array = _ellipse(array, [0, -0.22], [0.41, 0.16], 180 - 18, csf)
    # 6 CSF
    array = _ellipse(array, [0, 0.22], [0.31, 0.11], 180 + 18, csf)
    # 7 Tumor
    array = _ellipse(array, [0.35, 0], [0.21, 0.25], 0, tumor)
    # 8 Tumor
    array = _ellipse(array, [0.1, 0], [0.046, 0.046], 0, tumor)
    # 9 Tumor
    array = _ellipse(array, [-0.605, -0.08], [0.046, 0.023], 90, tumor)
    # 10 Tumor
    array = _ellipse(array, [-0.605, 0.06], [0.023, 0.046], 0, tumor)
    # 11 Tumor
    array = _ellipse(array, [-0.1, 0.0], [0.046, 0.046], 0, tumor)
    # 12 Tumor
    array = _ellipse(array, [-0.605, 0.0], [0.023, 0.023], 0, tumor)
    # 13 Sagittal sinus
    array = _ellipse(array, [-0.83, 0], [0.05, 0.05], 0, blood)
    # 14 Anterior cerebral artery
    array = _ellipse(array, [0.66, 0], [0.02, 0.02], 0, blood)
    # Flip
    array = np.flip(array, axis=0)
    return array


def shepp_logan(*params, n=256, B0=3):
    """Modified Shepp-Logan phantom mimicking an axial slice through the brain.

    The phantom is based on an MRI adaptation of the Shepp-Logan phantom 
    (Gach et al 2008), but with added features for use in a DC-MRI setting: 
    (1) additional regions for anterior cerebral artery and sinus sagittalis; 
    (2) additional optional contrasts blood flow (BF), blood volume (BV), 
    permeability-surface area product (PS) and interstitial volume (IV).

    Args:
        params (str or tuple): parameter or parameters shown in the image. 
          The options are 'PD' (proton density), 'T1', 'T2', 'Fb' (blood flow), 
          'vb' (Blood volume), 'PS' (permeability-surface area product) and 
          'vi' (interstitial volume). If no parameters are provided, the 
          function returns a dictionary with 14 masks, one for each region.
        n (int, optional): matrix size. Defaults to 256.
        B0 (int, optional): field strength in T. Defaults to 3.

    Reference:
        H. M. Gach, C. Tanase and F. Boada, "2D & 3D Shepp-Logan Phantom 
        Standards for MRI," 2008 19th International Conference on Systems 
        Engineering, Las Vegas, NV, USA, 2008, pp. 521-526, 
        `doi 10.1109/ICSEng.2008.15 <https://ieeexplore.ieee.org/document/4616690>`_.

    Returns:
        numpy.array or dict: if only one parameter is provided, this returns 
        an array. In all other conditions this returns a dictionary where keys 
        are the parameter- or region names, and values are square arrays with 
        image values.

    Note:
        Mask names:
            - background
            - scalp
            - bone
            - CSF skull
            - CSF left
            - CSF right
            - gray matter
            - tumor 1 to tumor 6
            - sagittal sinus
            - anterior artery

    Example:

    Generate a single contrast:

    .. plot::
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> import dcmri as dc

        Simulate a synthetic blood flow image:

        >>> im = dc.shepp_logan('Fb')

        Plot the result in units of mL/min/100mL:

        >>> fig, ax = plt.subplots(figsize=(5, 5), ncols=1)
        >>> pos = ax.imshow(6000*im, cmap='gray', vmin=0.0, vmax=80)
        >>> fig.colorbar(pos, ax=ax, label='blood flow (mL/min/100mL)')
        >>> plt.show()

    Generate multiple contrasts in one function call:

    .. plot::
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> import dcmri as dc

        Generate the MR Shepp-Logan phantom in low resolution:

        >>> im = dc.shepp_logan('PD', 'T1', 'T2', n=64)

        Plot the result:

        >>> fig, (ax1, ax2, ax3) = plt.subplots(figsize=(12, 5), ncols=3)
        >>> ax1.imshow(im['PD'], cmap='gray')
        >>> ax2.imshow(im['T1'], cmap='gray')
        >>> ax3.imshow(im['T2'], cmap='gray')
        >>> plt.show()

    Generate masks for the different regions-of-interest:

    .. plot::
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> import dcmri as dc

        Generate the MR Shepp-Logan phantom masks:

        >>> im = dc.shepp_logan(n=128)

        Plot all masks:

        >>> fig, ax = plt.subplots(figsize=(8, 8), ncols=4, nrows=4)
        >>> ax[0,0].imshow(im['background'], cmap='gray')
        >>> ax[0,1].imshow(im['scalp'], cmap='gray')
        >>> ax[0,2].imshow(im['bone'], cmap='gray')
        >>> ax[0,3].imshow(im['CSF skull'], cmap='gray')
        >>> ax[1,0].imshow(im['CSF left'], cmap='gray')
        >>> ax[1,1].imshow(im['CSF right'], cmap='gray')
        >>> ax[1,2].imshow(im['gray matter'], cmap='gray')
        >>> ax[1,3].imshow(im['tumor 1'], cmap='gray')
        >>> ax[2,0].imshow(im['tumor 2'], cmap='gray')
        >>> ax[2,1].imshow(im['tumor 3'], cmap='gray')
        >>> ax[2,2].imshow(im['tumor 4'], cmap='gray')
        >>> ax[2,3].imshow(im['tumor 5'], cmap='gray')
        >>> ax[3,0].imshow(im['tumor 6'], cmap='gray')
        >>> ax[3,1].imshow(im['sagittal sinus'], cmap='gray')
        >>> ax[3,2].imshow(im['anterior artery'], cmap='gray')
        >>> for i in range(4):
        >>>     for j in range(4):
        >>>         ax[i,j].set_yticklabels([])
        >>>         ax[i,j].set_xticklabels([])
        >>> plt.show()
    """
    if len(params) == 0:
        return _shepp_logan_mask(n=n)
    elif len(params) == 1:
        return _shepp_logan(params[0], n=n, B0=B0)
    else:
        phantom = {}
        for p in params:
            phantom[p] = _shepp_logan(p, n=n, B0=B0)
        return phantom



import os
import shutil
from copy import deepcopy
import numpy as np
import dcmri as dc



# Debugging mode
VERBOSE = 1
SHOW = True

VERBOSE = 0
SHOW = False




def test_ui_aorta_liver():

    time, aif, roi, gt = dc.fake_tissue()
    xdata, ydata = (time,time), (aif,roi)

    model = dc.AortaLiver(
        dt = 0.5,
        tmax = 180,
        weight = 70,
        agent = 'gadodiamide',
        dose = 0.2,
        rate = 3,
        field_strength = 3.0,
        t0 = 10,
        TR = 0.005,
        FA = 15,
        TS = 0.5,
        Th = 120,
    )
    model.free['Th'] = [0, np.inf]
    model.train(xdata, ydata, xtol=1e-3)
    model.plot(xdata, ydata, show=SHOW)
    assert model.cost(xdata, ydata) < 10
    assert 75 < model.params('Th', round_to=0) < 85


def test_ui_aorta_liver2scan():

    time, aif, roi, gt = dc.fake_tissue2scan(R10 = 1/dc.T1(3.0,'liver'))
    xdata = (time[0], time[1], time[0], time[1])
    ydata = (aif[0], aif[1], roi[0], roi[1])

    model = dc.AortaLiver2scan(
        dt = 0.5,
        weight = 70,
        agent = 'gadodiamide',
        dose = 0.2,
        dose2 = 0.2,
        rate = 3,
        field_strength = 3.0,
        t0 = 10,
        TR = 0.005,
        FA = 15,
        TS = 0.5,
        Th_i = 120,
        Th_f = 120,
    )
    model.free['Th_i'] = [0, np.inf]
    model.free['Th_f'] = [0, np.inf]
    model.train(xdata, ydata, xtol=1e-3)
    model.plot(xdata, ydata, show=SHOW)
    assert model.cost(xdata, ydata) < 5
    assert 60 < model.params('Th', round_to=0) < 80

def test_ui_liver():
    time, aif, vif, roi, gt = dc.fake_liver()

    # Show dual-inlet model
    params = {
        'kinetics': '2I-IC',
        'aif': aif,
        'vif': vif,
        'dt': time[1],
        'H': 0.45,
        'field_strength': 3,
        'agent': 'gadoxetate',
        'TR': 0.005,
        'FA': 15,
        'n0': 10,
        'R10': 1/dc.T1(3.0,'liver'),
        'R10a': 1/dc.T1(3.0, 'blood'),  
        'R10v': 1/dc.T1(3.0, 'blood'),      
    }
    model = dc.Liver(**params)
    model.train(time, roi)
    model.plot(time, roi, ref=gt, show=SHOW)
    assert model.cost(time, roi) < 0.1

    # Show single-inlet model
    params = {
        'kinetics': '1I-IC-D',
        'aif': aif,
        'dt': time[1],
        'H': 0.45,
        'field_strength': 3,
        'agent': 'gadoxetate',
        'TR': 0.005,
        'FA': 15,
        'n0': 10,
        'R10': 1/dc.T1(3.0,'liver'),
        'R10a': 1/dc.T1(3.0, 'blood'),        
    }
    model = dc.Liver(**params)
    model.train(time, roi)
    model.plot(time, roi, ref=gt, show=SHOW)
    assert model.cost(time, roi) < 1.5

    # Loop over all models
    for k in ['2I-EC', '2I-EC-HF', '1I-EC', '1I-EC-D', 
              '2I-IC', '2I-IC-HF', '2I-IC-U', '1I-IC-HF', 
              '1I-IC-D', '1I-IC-DU']:
        if k not in ['2I-IC-U', '1I-IC-DU']:
            stat = ['UE','U','E', None]
        else:
            stat = ['U', None]
        for s in stat:
            params_mdl = deepcopy(params)
            if k[0]=='2':
                params_mdl['vif'] = vif
            params_mdl['stationary'] = s
            params_mdl['kinetics'] = k
            model = dc.Liver(**params_mdl)
            model.train(time, roi, xtol=1e-2)
            # print(k, s, model.cost(time, roi))
            assert model.cost(time, roi) < 25

    # Display last result
    model.plot(time, roi, ref=gt, show=SHOW)


def test_ui_aorta_portal_liver():

    time, aif, vif, roi, gt = dc.fake_liver(sequence='SSI')
    xdata, ydata = (time, time, time), (aif, vif, roi)

    model = dc.AortaPortalLiver(
        sequence = 'SSI',
        dt = 0.5,
        tmax = 180,
        weight = 70,
        agent = 'gadoxetate',
        dose = 0.2,
        rate = 3,
        field_strength = 3.0,
        t0 = 10,
        TR = 0.005,
        FA = 15,
        TS = 0.5,
    )
    model.train(xdata, ydata, xtol=1e-3)
    model.plot(xdata, ydata, show=SHOW)
    model.print_params(round_to=3)
    assert 4 < model.cost(xdata, ydata) < 5



if __name__ == "__main__":

    # make_tmp()

    test_ui_aorta_liver()
    test_ui_aorta_liver2scan()
    test_ui_liver()
    test_ui_aorta_portal_liver()

    print('All mods tests passed!!')


import numpy as np
from scipy.integrate import trapezoid

import dcmri as dc
import dcmri.pk as pk

import matplotlib.pyplot as plt

def test_res_trap():
    t = np.linspace(0, 100, 20)
    r = dc.res_trap(t)
    assert np.sum(r) == 20

def test_prop_trap():
    t = np.linspace(0, 100, 20)
    h = dc.prop_trap(t)
    assert np.linalg.norm(h) == 0

def test_conc_trap():
    t = np.linspace(0, 100, 20)
    J = np.ones(len(t))
    C = dc.conc(J, t=t, model='trap')
    assert C[-1] == 100
    C = dc.conc(J, dt=t[1], model='trap')
    assert C[-1] == 100
    C0 = dc.conv(dc.res_trap(t), J, t)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) == 0

def test_flux_trap():
    t = np.linspace(0, 100, 20)
    J = np.ones(len(t))
    Jo = dc.flux(J, model='trap')
    assert np.linalg.norm(Jo) == 0
    Jo0 = dc.conv(dc.prop_trap(t), J, t)
    assert np.linalg.norm(Jo-Jo0) == 0

def test_res_pass():
    T = 30
    t = np.linspace(0, 100, 20)
    r = dc.res_pass(T, t)
    assert round(r[0] * t[1]/2) == T
    assert np.linalg.norm(r[1:]) == 0
    
def test_prop_pass():
    T = 30
    t = np.linspace(0, 100, 20)
    h = dc.prop_pass(t)
    assert round(h[0] * t[1]/2) == 1
    assert np.linalg.norm(h[1:]) == 0

def test_conc_pass():
    T = 30
    t = np.linspace(0, 100, 20)
    J = np.ones(len(t))
    C = dc.conc(J, T, model='pass')
    assert np.unique(C)[0] == T
    C0 = dc.conv(dc.res_pass(T,t), J, t)
    assert np.linalg.norm(C[1:]-C0[1:])/np.linalg.norm(C0[1:]) < 1e-12

def test_flux_pass():
    T = 30
    t = np.linspace(0, 100, 20)
    J = np.ones(len(t))
    Jo = dc.flux(J, model='pass')
    assert np.linalg.norm(J-Jo)/np.linalg.norm(J) < 1e-12
    Jo0 = dc.conv(dc.prop_pass(t), J, t)
    assert np.linalg.norm(Jo[1:]-Jo0[1:])/np.linalg.norm(Jo[1:]) < 1e-12


def test_res_comp():
    T = 25
    t = np.linspace(0, 150, 20)
    r = dc.res_comp(T, t)
    assert np.round(trapezoid(r,t)) == T
    t = [0,5,15,30,60,90,150]
    r = dc.res_comp(T, t)
    assert (trapezoid(r,t)-T)**2/T**2 < 1e-2
    r = dc.res_comp(np.inf, t)
    assert np.sum(np.unique(r))== 1
    r = dc.res_comp(0, t)
    assert np.sum(r) == 1

def test_prop_comp():
    T = 25
    t = np.linspace(0, 150, 20)
    h = dc.prop_comp(T, t)
    assert np.round(trapezoid(h,t)) == 1
    t = [0,5,15,30,60,90,150]
    h = dc.prop_comp(T, t)
    assert (trapezoid(h,t)-1)**2 < 1e-2
    h = dc.prop_comp(np.inf, t)
    assert np.linalg.norm(h)==0
    h = dc.prop_comp(0, t)
    assert (trapezoid(h,t)-1)**2 < 1e-2

def test_conc_comp():
    T = 25
    t = np.linspace(0, 150, 20)
    J = np.ones(len(t))
    C0 = dc.conv(dc.res_comp(T,t), J, t)
    C = dc.conc(J, T, t=t, model='comp')
    assert np.linalg.norm(C[1:]-C0[1:])/np.linalg.norm(C0[1:]) < 1e-2
    C = dc.conc(J, T, dt=t[1], model='comp')
    assert np.linalg.norm(C[1:]-C0[1:])/np.linalg.norm(C0[1:]) < 1e-2
    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    C0 = dc.conv(dc.res_comp(T,t), J, t)
    C = dc.conc(J, T, t=t, model='comp')
    assert np.linalg.norm(C[1:]-C0[1:])/np.linalg.norm(C0[1:]) < 1e-1
    C = dc.conc(J, np.inf, t=t, model='comp')
    C0 = dc.conc(J,t=t, model='trap')
    assert np.linalg.norm(C[1:]-C0[1:])/np.linalg.norm(C0[1:]) < 1e-1

def test_flux_comp():
    T = 25
    t = np.linspace(0, 150, 20)
    J = np.ones(len(t))
    J0 = dc.conv(dc.prop_comp(T,t), J, t)
    Jo = dc.flux(J, T, t=t, model='comp')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 1e-2
    Jo = dc.flux(J, T, dt=t[1], model='comp')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 1e-2
    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    J0 = dc.conv(dc.prop_comp(T,t), J, t)
    Jo = dc.flux(J, T, t=t, model='comp')
    assert np.linalg.norm(J0-Jo)/np.linalg.norm(J0) < 1e-1
    Jo = dc.flux(J, np.inf, t=t, model='comp')
    assert np.linalg.norm(Jo) == 0

def test_res_plug():
    T = 25
    t = np.linspace(0, 150, 20)
    r = dc.res_plug(T, t)
    assert (trapezoid(r,t)-T)**2/T**2 < 0.02
    t = [0,5,15,30,60,90,150]
    r = dc.res_plug(T, t)
    assert (trapezoid(r,t)-T)**2/T**2 < 0.06

def test_prop_plug():
    T = 25
    t = np.linspace(0, 150, 500)
    h = dc.prop_plug(T, t)
    assert np.abs((trapezoid(h,t)-1)) < 1e-12
    t = [0,5,15,30,60,90,150]
    h = dc.prop_plug(T, t)
    assert np.abs((trapezoid(h,t)-1)) < 1e-12

def test_conc_plug():
    T = 25
    t = np.linspace(0, 150, 40)
    J = np.ones(len(t))
    C0 = dc.conv(dc.res_plug(T,t), J, t)
    C = dc.conc(J, T, t=t, model='plug')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-2
    C = dc.conc_plug(J, T, t, solver='conv')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-2
    C = dc.conc_plug(J, T, dt=t[1])
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-2
    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    C0 = dc.conv(dc.res_plug(T,t), J, t)
    C = dc.conc_plug(J, T, t)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1
    C = dc.conc_plug(J, np.inf, t)
    C0 = dc.conc_trap(J)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-2
    C = dc.conc_plug(J, 0, t)
    assert np.linalg.norm(C) == 0

def test_flux_plug():
    T = 25
    t = np.linspace(0, 150, 40)
    J = np.ones(len(t))
    J0 = dc.conv(dc.prop_plug(T,t), J, t)
    Jo = dc.flux(J, T, t=t, model='plug')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 0.2
    Jo = dc.flux(J, T, t=t, model='plug', solver='conv')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 0.2
    Jo = dc.flux(J, T, dt=t[1], model='plug')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 0.2
    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    J0 = dc.conv(dc.prop_plug(T,t), J, t)
    Jo = dc.flux(J, T, t=t, model='plug')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 0.3
    Jo = dc.flux(J, np.inf, t=t, model='plug')
    assert np.linalg.norm(Jo) == 0
    Jo = dc.flux(J, 0, t=t, model='plug')
    assert np.linalg.norm(Jo-J) == 0

    try:
        Jo = dc.flux(J, T, t=t, model='plug', solver='blabla')
    except:
        assert True
    else:
        assert False


def test_prop_chain():
    T = 25
    D = 0.5
    t = np.linspace(0, 150, 500)
    h = dc.prop_chain(T, D, t)
    assert np.abs(trapezoid(h,t)-1) < 0.001
    assert np.abs(trapezoid(t*h,t)-T) < 0.02
    t = [0,5,15,30,60,90,150]
    h = dc.prop_chain(T, D, t)
    assert np.abs(trapezoid(h,t)-1) < 0.03
    assert np.abs(trapezoid(t*h,t)-T) < 0.5
    try:
        h = dc.prop_chain(-1, D, t)
    except:
        assert True
    else:
        assert False
    try:
        h = dc.prop_chain(T, -1, t)
    except:
        assert True
    else:
        assert False
    try:
        h = dc.prop_chain(T, 2, t)
    except:
        assert True
    else:
        assert False
    h = dc.prop_chain(T, 0, t)
    h0 = dc.prop_plug(T,t)
    assert np.linalg.norm(h-h0)==0
    h = dc.prop_chain(T, 1, t)
    h0 = dc.prop_comp(T, t)
    assert np.linalg.norm(h-h0)==0


def test_res_chain():
    T = 25
    D = 0.5
    t = np.linspace(0, 150, 500)
    h = dc.res_chain(T, D, t)
    assert (trapezoid(h,t)-T)**2/T**2 < 1e-6
    t = [0,5,15,30,60,90,150]
    h = dc.res_chain(T, D, t)
    assert (trapezoid(h,t)-T)**2/T**2 < 1e-3
    h = dc.res_chain(T, 0, t)
    h0 = dc.res_plug(T,t)
    assert np.linalg.norm(h-h0)==0
    h = dc.res_chain(T, 1, t)
    h0 = dc.res_comp(T, t)
    assert np.linalg.norm(h-h0)==0

def test_conc_chain():
    T = 25
    D = 0.5
    t = np.linspace(0, 150, 20)
    J = np.ones(len(t))
    C0 = dc.conv(dc.res_chain(T,D,t), J, t)
    C = dc.conc(J, T, D, t=t, model='chain')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-12
    C = dc.conc_chain(J, T, D, dt=t[1])
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-12
    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    C0 = dc.conv(dc.res_chain(T,D,t), J, t)
    C = dc.conc_chain(J, T, D, t)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-12
    h = dc.conc_chain(J, T, 0, t)
    h0 = dc.conc_plug(J, T,t)
    assert np.linalg.norm(h-h0)==0
    h = dc.conc_chain(J, T, 1, t)
    h0 = dc.conc_comp(J, T, t)
    assert np.linalg.norm(h-h0)==0

def test_flux_chain():
    T = 25
    D = 0.5
    t = np.linspace(0, 150, 20)
    J = np.ones(len(t))
    J0 = dc.conv(dc.prop_chain(T,D,t), J, t)
    Jo = dc.flux(J, T, D, t=t, model='chain')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 1e-12
    Jo = dc.flux(J, T, D, dt=t[1], model='chain')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 1e-12
    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    J0 = dc.conv(dc.prop_chain(T,D,t), J, t)
    Jo = dc.flux(J, T, D, t=t, model='chain')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 1e-12
    h = dc.flux(J, T, 0, t=t, model='chain')
    h0 = dc.flux(J, T, t=t, model='plug')
    assert np.linalg.norm(h-h0)==0
    h = dc.flux(J, T, 1, t=t, model='chain')
    h0 = dc.flux(J, T, t=t, model='comp')
    assert np.linalg.norm(h-h0)==0



def test_flux_pfcomp():
    T = 10
    D = 0.5
    t = np.linspace(0, 300, 1000)
    J = np.exp(-t/20)/20

    # Check that the function preserves areas
    Jo = dc.flux(J, T, D, t=t, model='pfcomp')
    assert np.sum(Jo)*t[1]-1 < 1e-2

    # D=0
    Jo = dc.flux(J, T, 0, t=t, model='pfcomp')
    assert np.sum(Jo)*t[1]-1 < 1e-2

    # D=1
    Jo = dc.flux(J, T, 1, t=t, model='pfcomp')
    assert np.sum(Jo)*t[1]-1 < 1e-2

    try:
        Jo = dc.flux(J, T, -1, t=t, model='pfcomp')
    except:
        assert True
    else:
        assert False
    try:
        Jo = dc.flux(J, T, 2, t=t, model='pfcomp')
    except:
        assert True
    else:
        assert False


def test_prop_step():
    T, D = 25, 0.5
    t = np.linspace(0, 150, 100)
    h = dc.prop_step(T, D, t)
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    assert np.abs(trapezoid(t*h,t)-T) < 1e-12
    h = dc.prop_step(T, 1, t)
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    h = dc.prop_step(120, 0.5, t)
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    h = dc.prop_step(120, 0.5, t)
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    t = [0,5,15,30,60,90,150]
    h = dc.prop_step(T, D, t)
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    try:
        h = dc.prop_step(-1, D, t)
    except:
        assert True
    else:
        assert False
    try:
        h = dc.prop_step(T, -1, t)
    except:
        assert True
    else:
        assert False
    try:
        h = dc.prop_step(T, 2, t)
    except:
        assert True
    else:
        assert False
    h = dc.prop_step(np.inf, D, t)
    assert np.linalg.norm(h)==0
    h = dc.prop_step(T, 0, t)
    h0 = dc.prop_plug(T, t)
    assert np.linalg.norm(h-h0)==0

def test_res_step():
    T, D = 25, 0.5
    t = np.linspace(0, 150, 500)
    h = dc.res_step(T, D, t)
    assert (trapezoid(h,t)-T)**2/T**2 < 1e-9
    t = [0,5,15,30,60,90,150]
    h = dc.res_step(T, D, t)
    assert (trapezoid(h,t)-T)**2/T**2 < 0.5

def test_conc_step():
    T, D = 25, 0.5
    t = np.linspace(0, 150, 20)
    J = np.ones(len(t))
    C0 = dc.conv(dc.res_step(T,D,t), J, t)
    C = dc.conc(J, T, D, t=t, model='step')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-12
    C = dc.conc_step(J, T, D, dt=t[1])
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-12
    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    C0 = dc.conv(dc.res_step(T,D,t), J, t)
    C = dc.conc_step(J, T, D, t)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-12
    C = dc.conc_step(J, T, 0, t)
    C0 = dc.conc_plug(J, T, t)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-12


def test_flux_step():
    T, D = 25, 0.5
    t = np.linspace(0, 150, 20)
    J = np.ones(len(t))
    J0 = dc.conv(dc.prop_step(T,D,t), J, t)
    Jo = dc.flux(J, T, D, t=t, model='step')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 1e-12
    Jo = dc.flux(J, T, D, dt=t[1], model='step')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 1e-12
    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    J0 = dc.conv(dc.prop_step(T,D,t), J, t)
    Jo = dc.flux(J, T, D, t=t, model='step')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 1e-12
    Jo = dc.flux(J, T, 0, t=t, model='step')
    J0 = dc.flux(J, T, t=t, model='plug')
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 1e-12


def test_prop_free():
    t = np.linspace(0, 150, 100)
    h = dc.prop_free([1,1], t, TT=[30,60,90])
    T = 60
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    assert (trapezoid(t*h,t)-T)**2/T**2 < 1e-9
    h = dc.prop_free([2,2], t, TT=[30,60,90])
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    assert (trapezoid(t*h,t)-T)**2/T**2 < 1e-9
    h = dc.prop_free([2,5], t, TT=[30,60,90])
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    h = dc.prop_free([2,5], t, TT=[30,60,120])
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    h = dc.prop_free([2,5], t, TT=[30.5,60.5,120.5])
    assert np.abs(trapezoid(h,t)-1) < 1e-12
    h = dc.prop_free([1,1], t)
    h0 = dc.prop_free([1,1], t, TTmax=np.amax(t))
    assert np.linalg.norm(h-h0)/np.linalg.norm(h0) < 1e-12
    try:
        h = dc.prop_free([1,1,1], t, TT=[30,60,90])
    except:
        assert True
    else:
        assert False


def test_res_free():
    t = np.linspace(0, 150, 100)
    h = dc.res_free([1,1], t, TT=[30,60,90])
    T = 60
    assert (trapezoid(h,t)-T)**2/T**2 < 1e-9

def test_conc_free():
    H = [1,1]
    TT = [30,60,90]
    t = np.linspace(0, 150, 20)
    J = np.ones(len(t))
    C0 = dc.conv(dc.res_free(H,t,TT), J, t)
    C = dc.conc(J, H, t=t, model='free', TT=TT)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-12
    C = dc.conc_free(J, H, t, dt=t[1], TT=TT)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-12
    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    C0 = dc.conv(dc.res_free(H,t,TT=TT), J, t)
    C = dc.conc_free(J, H, t, TT=TT)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-12

def test_flux_free():
    H = [1,1]
    TT = [30,60,90]
    t = np.linspace(0, 150, 20)
    J = np.ones(len(t))
    J0 = dc.conv(dc.prop_free(H,t,TT), J, t)
    Jo = dc.flux(J, H, t=t, model='free', TT=TT)
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 1e-12
    Jo = dc.flux(J, H, dt=t[1], model='free', TT=TT)
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 1e-12
    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    J0 = dc.conv(dc.prop_free(H,t,TT=TT), J, t)
    Jo = dc.flux(J, H, t=t, model='free', TT=TT)
    assert np.linalg.norm(Jo-J0)/np.linalg.norm(J0) < 1e-12


def test__K_ncomp():
    T1, T2 = 2, 7
    E11, E12, E21, E22 = 1, 3, 9, 2
    K = pk._K_ncomp([T1,T2], [[E11,E12],[E21,E22]])
    K0 = [[(E11+E21)/T1, -E12/T2],[-E21/T1, (E12+E22)/T2]]
    assert np.array_equal(K, K0)
    assert np.array_equal(np.matmul(K, [1,0]), [(E11+E21)/T1,-E21/T1])
    assert np.array_equal(np.matmul(K, [0,1]), [-E12/T2,(E12+E22)/T2])
    assert np.array_equal(np.matmul(K, [1,0]), K[:,0]) # col 0
    assert np.array_equal(np.matmul(K, [0,1]), K[:,1]) # col 1
    try:
        pk._K_ncomp([T1,T2], [[E11,-1],[E21,E22]])
    except:
        assert True
    else:
        assert False
    K = pk._K_ncomp([T1,T2], [[0,E12],[0,E22]])
    assert K[0,0] == 0

def test__J_ncomp():
    T = [1,2]
    E = [[1,0],[0,1]]
    C = [[1],
         [3]]
    J = pk._J_ncomp(np.array(C), T, E)
    assert J[0,0,:] == 1
    assert J[0,1,:] == 0
    assert J[1,0,:] == 0
    assert J[1,1,:] == 1.5

def test_conc_ncomp_prop():
    # Compare a decoupled system against analytical 1-comp model solutions
    t = np.linspace(0, 20, 10)
    J = np.ones((2,len(t)))
    J[1,:] = 2
    T = [6,12]
    E = [[1,0],[0,1]]
    C0 = dc.conc_comp(J[0,:], T[0], t)
    C1 = dc.conc_comp(J[1,:], T[1], t)
    prec = [1e-1, 1e-2, 1e-3]
    for i, dt_prop in enumerate([None, 0.1, 0.01]):
        C = dc.conc_ncomp_prop(J, T, E, t, dt_prop=dt_prop)
        assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < prec[i]
        assert np.linalg.norm(C[1,:]-C1)/np.linalg.norm(C1) < prec[i]

def test_conc_ncomp_diag():
    # Compare a decoupled system against analytical 1-comp model solutions
    t = np.linspace(0, 20, 10)
    J = np.ones((2,len(t)))
    J[1,:] = 2
    T = [6,12]
    E = [[1,0],[0,1]]
    C0 = dc.conc_comp(J[0,:], T[0], t)
    C1 = dc.conc_comp(J[1,:], T[1], t)
    C = dc.conc_ncomp_diag(J, T, E, t)
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 1e-6
    assert np.linalg.norm(C[1,:]-C1)/np.linalg.norm(C1) < 1e-6

def test_conc_ncomp():

    # Compare a decoupled system against analytical 1-comp model solutions
    t = np.linspace(0, 20, 10)
    J = np.ones((2,len(t)))
    J[1,:] = 2
    T = [6,12]
    E = [[1,0],[0,1]]
    C0 = dc.conc_comp(J[0,:], T[0], t)
    C1 = dc.conc_comp(J[1,:], T[1], t)
    prec = [1e-1, 1e-2, 1e-3]
    for i, dt_prop in enumerate([None, 0.1, 0.01]):
        C = dc.conc(J, T, E, t=t, model='ncomp', dt_prop=dt_prop)
        assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < prec[i]
        assert np.linalg.norm(C[1,:]-C1)/np.linalg.norm(C1) < prec[i]

    # Compare both solvers for a few coupled systems
    E = [[1,0.5],[0.5,1]]
    C = dc.conc_ncomp(J, T, E, t, solver='prop', dt_prop=0.01)
    C0 = dc.conc_ncomp(J, T, E, t, solver='diag')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C) < 1e-3
    E = [[0.2,0.6],[0.8,0.4]]
    C = dc.conc_ncomp(J, T, E, t, solver='prop', dt_prop=0.01)
    C0 = dc.conc_ncomp(J, T, E, t, solver='diag')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C) < 1e-3
    E = [[0.6, 0.5, 0.0], [0.1, 0.2, 0.5], [0.3, 0.3, 0.5]]
    T3 = [6,12,6]
    J3 = np.ones((3,len(t)))
    C = dc.conc_ncomp(J3, T3, E, t, solver='prop', dt_prop=0.01)
    C0 = dc.conc_ncomp(J3, T3, E, t, solver='diag')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C) < 1e-1

    # Compare both solvers for a 2CXM
    J = np.ones(len(t))
    J = np.stack((J, np.zeros(J.size)))
    E = 0.1
    E = [[1-E,1],[E,0]]
    C = dc.conc(J, T, E, t=t, model='ncomp', solver='prop', dt_prop=0.01)
    C0 = dc.conc(J, T, E, t=t, model='ncomp', solver='diag')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C) < 1e-3

def test_flux_ncomp():
    # Compare a decoupled system against analytical 1-comp model solutions
    t = np.linspace(0, 20, 10)
    J = np.ones((2,len(t)))
    J[1,:] = 2
    T = [6,12]
    E = [[1,0],[0,1]]
    J0 = dc.conc_comp(J[0,:], T[0], t) / T[0]
    J1 = dc.conc_comp(J[1,:], T[1], t) / T[1]
    J = dc.flux(J, T, E, t=t, dt_prop=0.01, model='ncomp')
    assert np.linalg.norm(J[0,0,:]-J0)/np.linalg.norm(J0) < 1e-3
    assert np.linalg.norm(J[1,1,:]-J1)/np.linalg.norm(J1) < 1e-3
    assert np.linalg.norm(J[1,0,:]) == 0
    assert np.linalg.norm(J[0,1,:]) == 0

def test_res_ncomp():
    # Compare a decoupled system against analytical 1-comp model solutions.
    t = np.linspace(0, 20, 10)
    T = [6,12]
    E = [[1,0],[0,1]]
    R0 = dc.res_comp(T[0], t)
    R1 = dc.res_comp(T[1], t)
    R = dc.res_ncomp(T, E, t)
    assert np.linalg.norm(R[0,0,:]-R0)/np.linalg.norm(R0) < 1e-12
    assert np.linalg.norm(R[1,1,:]-R1)/np.linalg.norm(R1) < 1e-12
    assert np.linalg.norm(R[1,0,:]) == 0
    assert np.linalg.norm(R[0,1,:]) == 0
    # Compare against convolution on a coupled system.
    E = [[0.6, 0.5, 0.0], [0.1, 0.2, 0.5], [0.3, 0.3, 0.5]]
    T = [6, 12, 6]
    J = np.ones((3,len(t)))
    C = dc.conc_ncomp(J, T, E, t, solver='prop', dt_prop=0.01)
    R = dc.res_ncomp(T, E, t)
    C0 = dc.conv(R[0,0,:], J[0,:], t) + dc.conv(R[1,0,:], J[1,:], t)
    C1 = dc.conv(R[0,1,:], J[0,:], t) + dc.conv(R[1,1,:], J[1,:], t)
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 0.1
    assert np.linalg.norm(C[1,:]-C1)/np.linalg.norm(C1) < 0.5

def test_prop_ncomp():
    # Compare a decoupled system against analytical 1-comp model solutions
    t = np.linspace(0, 20, 10)
    T = [6,12]
    E = [[1,0],[0,1]]
    H0 = dc.prop_comp(T[0], t) 
    H1 = dc.prop_comp(T[1], t)
    H = dc.prop_ncomp(T, E, t)
    assert np.linalg.norm(H[0,0,0,:]-H0)/np.linalg.norm(H0) < 1e-12
    assert np.linalg.norm(H[1,1,1,:]-H1)/np.linalg.norm(H1) < 1e-12
    assert np.linalg.norm(H[:,1,0,:]) == 0
    assert np.linalg.norm(H[:,0,1,:]) == 0
    # Compare against convolution on a coupled system
    J = np.ones((2,len(t)))
    J[1,:] = 2
    E = [[1,0.5],[0.5,1]]
    Jo = dc.flux_ncomp(J, T, E, t, solver='prop', dt_prop=0.01)
    H = dc.prop_ncomp(T, E, t)
    Jo0 = dc.conv(H[0,0,0,:], J[0,:], t) + dc.conv(H[1,0,0,:], J[1,:], t)
    Jo1 = dc.conv(H[0,1,1,:], J[0,:], t) + dc.conv(H[1,1,1,:], J[1,:], t)
    assert np.linalg.norm(Jo[0,0,:]-Jo0)/np.linalg.norm(Jo0) < 0.1
    assert np.linalg.norm(Jo[1,1,:]-Jo1)/np.linalg.norm(Jo1) < 0.01


def test__K_2comp():
    # Compare against numerical computation
    T = [6,12]
    E = [[1,0.5],[0.5,1]]
    # Analytical
    _, K, _ = pk._K_2comp(T,E)
    # Numerical
    Knum = pk._K_ncomp(T, E)
    Knum, _ = np.linalg.eig(Knum)
    assert np.linalg.norm(K-Knum)/np.linalg.norm(Knum) < 1e-12

def test_conc_2comp():

    # Compare a coupled system against numerical solutions
    t = np.linspace(0, 20, 10)
    J = np.ones((2,len(t)))
    J[1,:] = 2
    T = [6,12]
    E = [[1,0.5],[0.5,1]]
    C0 = dc.conc(J, T, E, t=t, model='ncomp') 
    C = dc.conc_ncomp(J, T, E, t, dt_prop=0.1)
    assert np.linalg.norm(C[0,:]-C0[0,:])/np.linalg.norm(C[0,:]) < 1e-2
    assert np.linalg.norm(C[1,:]-C0[1,:])/np.linalg.norm(C[1,:]) < 1e-2

    # Compare a decoupled system against individual solutions
    E = [[1,0],[0,1]]
    C = dc.conc_ncomp(J, T, E, t)
    C0 = dc.conc_comp(J[0,:], T[0], t)
    C1 = dc.conc_comp(J[1,:], T[1], t)
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 1e-12
    assert np.linalg.norm(C[1,:]-C1)/np.linalg.norm(C1) < 1e-12

    # Test exceptions
    try:
        dc.conc_ncomp(J, [6,0], E, t)
    except:
        assert True
    else:
        assert False
    C = dc.conc_ncomp([[1,1],[1,1]], [1,1], [[1,0],[0,1]])
    assert C[0,0]==0


def test_flux_2comp():
    # Compare a decoupled system against analytical 1-comp model solutions
    t = np.linspace(0, 20, 10)
    J = np.ones((2,len(t)))
    J[1,:] = 2
    T = [6,12]
    E = [[1,0],[0,1]]
    J0 = dc.conc_comp(J[0,:], T[0], t) / T[0]
    J1 = dc.conc_comp(J[1,:], T[1], t) / T[1]
    J = dc.flux(J, T, E, t=t, model='ncomp')
    assert np.linalg.norm(J[0,0,:]-J0)/np.linalg.norm(J0) < 1e-3
    assert np.linalg.norm(J[1,1,:]-J1)/np.linalg.norm(J1) < 1e-3
    assert np.linalg.norm(J[1,0,:]) == 0
    assert np.linalg.norm(J[0,1,:]) == 0

def test_res_2comp():
    # Compare a decoupled system against analytical 1-comp model solutions.
    t = np.linspace(0, 20, 10)
    T = [6,12]
    E = [[1,0],[0,1]]
    R0 = dc.res_comp(T[0], t)
    R1 = dc.res_comp(T[1], t)
    R = dc.res_ncomp(T, E, t)
    assert np.linalg.norm(R[0,0,:]-R0)/np.linalg.norm(R0) < 1e-12
    assert np.linalg.norm(R[1,1,:]-R1)/np.linalg.norm(R1) < 1e-12
    assert np.linalg.norm(R[1,0,:]) == 0
    assert np.linalg.norm(R[0,1,:]) == 0
    # Compare against convolution on a coupled system.
    J = np.ones((2,len(t)))
    J[1,:] = 2
    E = [[1,0.5],[0.5,1]]
    C = dc.conc_ncomp(J, T, E, t)
    R = dc.res_ncomp(T, E, t)
    C0 = dc.conv(R[0,0,:], J[0,:], t) + dc.conv(R[1,0,:], J[1,:], t)
    C1 = dc.conv(R[0,1,:], J[0,:], t) + dc.conv(R[1,1,:], J[1,:], t)
    assert np.linalg.norm(C[0,:]-C0)/np.linalg.norm(C0) < 0.1
    assert np.linalg.norm(C[1,:]-C1)/np.linalg.norm(C1) < 0.01

def test_prop_2comp():
    # Compare a decoupled system against analytical 1-comp model solutions
    t = np.linspace(0, 20, 10)
    T = [6,12]
    E = [[1,0],[0,1]]
    H0 = dc.prop_comp(T[0], t) 
    H1 = dc.prop_comp(T[1], t)
    H = dc.prop_ncomp(T, E, t)
    assert np.linalg.norm(H[0,0,0,:]-H0)/np.linalg.norm(H0) < 1e-12
    assert np.linalg.norm(H[1,1,1,:]-H1)/np.linalg.norm(H1) < 1e-12
    assert np.linalg.norm(H[:,1,0,:]) == 0
    assert np.linalg.norm(H[:,0,1,:]) == 0
    # Compare against convolution on a coupled system
    J = np.ones((2,len(t)))
    J[1,:] = 2
    E = [[1,0.5],[0.5,1]]
    Jo = dc.flux_ncomp(J, T, E, t)
    H = dc.prop_ncomp(T, E, t)
    Jo0 = dc.conv(H[0,0,0,:], J[0,:], t) + dc.conv(H[1,0,0,:], J[1,:], t)
    Jo1 = dc.conv(H[0,1,1,:], J[0,:], t) + dc.conv(H[1,1,1,:], J[1,:], t)
    assert np.linalg.norm(Jo[0,0,:]-Jo0)/np.linalg.norm(Jo0) < 0.1
    assert np.linalg.norm(Jo[1,1,:]-Jo1)/np.linalg.norm(Jo1) < 0.01

def test_conc_nscomp():

    # Some inputs
    T = 25
    t = np.linspace(0, 150, 20)
    J = np.ones(len(t))

    # Test exceptions
    try:
        dc.conc_nscomp(J, T, t)
    except:
        assert True
    try:
        dc.conc_nscomp(J, [T,T], t)
    except:
        assert True
    try:
        dc.conc_nscomp(J, -T*np.ones(len(t)), t)
    except:
        assert True

    # Test functionality
    C = dc.conc_comp(J, T, t)
    C0 = dc.conc(J, T*np.ones(len(t)), t=t, model='nscomp')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.1
    
    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    C = dc.conc_comp(J, T, t)
    C0 = dc.conc_nscomp(J, T*np.ones(len(t)), t)
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.1


def test_flux_nscomp():

    # Some inputs
    T = 25
    t = np.linspace(0, 150, 20)
    J = np.ones(len(t))

    # Test functionality
    C = dc.flux(J, T, t=t, model='comp')
    C0 = dc.flux(J, T*np.ones(len(t)), t=t, model='nscomp')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.1
    
    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    C = dc.flux(J, T, t=t, model='comp')
    C0 = dc.flux(J, T*np.ones(len(t)), t=t, model='nscomp')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.1


def test_conc_mmcomp():
    T = 25
    t = np.linspace(0, 150, 500)
    J = np.ones(len(t))
    C = dc.conc_comp(J, T, t)
    Km = np.amax(C)/2
    Vmax = np.amax(C)/np.amax(T)
    C = dc.conc(J, Vmax, Km, t=t, model='mmcomp', solver='SM')
    C0 = dc.conc_mmcomp(J, Vmax, Km, t, solver='prop')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-2

    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    C = dc.conc_mmcomp(J, Vmax, Km, t, solver='SM')
    C0 = dc.conc_mmcomp(J, Vmax, Km, t, solver='prop')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.3

    try:
        C = dc.conc_mmcomp(J, -Vmax, Km, t, solver='SM')
    except:
        True
    try:
        C = dc.conc_mmcomp(J, Vmax, -Km, t, solver='SM')
    except:
        True

def test_flux_mmcomp():
    T = 25
    t = np.linspace(0, 150, 500)
    J = np.ones(len(t))
    C = dc.conc_comp(J, T, t)
    Km = np.amax(C)/2
    Vmax = np.amax(C)/np.amax(T)
    C = dc.flux(J, Vmax, Km, t=t, model='mmcomp', solver='SM')
    C0 = dc.flux(J, Vmax, Km, t=t, model='mmcomp', solver='prop')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 1e-2

    t = [0,5,15,30,60,90,150]
    J = np.ones(len(t))
    C = dc.flux(J, Vmax, Km, t=t, model='mmcomp', solver='SM')
    C0 = dc.flux(J, Vmax, Km, t=t, model='mmcomp', solver='prop')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C0) < 0.3

def test_conc_2cxm():
    # Test against general 2-compartment model
    t = np.linspace(0, 20, 10)
    J = np.ones(len(t))
    J2c = np.stack((J, np.zeros(J.size)))
    
    T = [6,12]
    E = 0.1
    C0 = dc.conc(J, T, E, t=t, model='2cxm')
    E = [[1-E,1],[E,0]]
    C = dc.conc(J2c, T, E, t=t, model='ncomp')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C) < 1e-9

    E = 0.5
    C0 = dc.conc(J, T, E, t=t, model='2cxm')
    E = [[1-E,1],[E,0]]
    C = dc.conc(J2c, T, E, t=t, model='ncomp')
    assert np.linalg.norm(C-C0)/np.linalg.norm(C) < 1e-9

    # # test boundary case: TP=0
    E = 0.5
    T = [6,12]
    C0 = dc.conc(J, T, E, t=t, model='2cxm')
    T[0] = 0.1
    C1 = dc.conc(J, T, E, t=t, model='2cxm')
    T[0] = 0.01
    C2 = dc.conc(J, T, E, t=t, model='2cxm')
    T[0] = 0.0
    C = dc.conc(J, T, E, t=t, model='2cxm')
    # Check convergence to T=0 solution
    err0 = np.linalg.norm(C0-C)
    err1 = np.linalg.norm(C1-C)
    err2 = np.linalg.norm(C2-C)
    assert err1 < err0
    assert err2 < err1
    assert err2/np.linalg.norm(C) < 1e-2

    # # test boundary case: TE=0
    E = 0.5
    T = [6,12]
    C0 = dc.conc(J, T, E, t=t, model='2cxm')
    T[1] = 0.1
    C1 = dc.conc(J, T, E, t=t, model='2cxm')
    T[1] = 0.01
    C2 = dc.conc(J, T, E, t=t, model='2cxm')
    T[1] = 0.0
    C = dc.conc(J, T, E, t=t, model='2cxm')
    # Check convergence to T=0 solution
    err0 = np.linalg.norm(C0-C)
    err1 = np.linalg.norm(C1-C)
    err2 = np.linalg.norm(C2-C)
    assert err1 < err0
    assert err2 < err1
    assert err2/np.linalg.norm(C) < 1e-2

    # # test boundary case: TP=inf
    E = 0.5
    T = [6,12]
    C0 = dc.conc(J, T, E, t=t, model='2cxm')
    T[0] = 100
    C1 = dc.conc(J, T, E, t=t, model='2cxm')
    T[0] = 1000
    C2 = dc.conc(J, T, E, t=t, model='2cxm')
    T[0] = np.inf
    C = dc.conc(J, T, E, t=t, model='2cxm')
    # Check convergence to T=0 solution
    err0 = np.linalg.norm(C0-C)
    err1 = np.linalg.norm(C1-C)
    err2 = np.linalg.norm(C2-C)
    assert err1 < err0
    assert err2 < err1
    assert err2/np.linalg.norm(C) < 1e-2

    # # test boundary case: TE=inf
    E = 0.5
    T = [6,12]
    C0 = dc.conc(J, T, E, t=t, model='2cxm')
    T[1] = 100
    C1 = dc.conc(J, T, E, t=t, model='2cxm')
    T[1] = 1000
    C2 = dc.conc(J, T, E, t=t, model='2cxm')
    T[1] = np.inf
    C = dc.conc(J, T, E, t=t, model='2cxm')
    # Check convergence to T=0 solution
    err0 = np.linalg.norm(C0-C)
    err1 = np.linalg.norm(C1-C)
    err2 = np.linalg.norm(C2-C)
    assert err1 < err0
    assert err2 < err1
    assert err2/np.linalg.norm(C) < 1e-2

    # # test boundary case: E=0
    E = 0.5
    T = [6,12]
    C0 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0.05
    C1 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0.005
    C2 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0
    C = dc.conc(J, T, E, t=t, model='2cxm')
    # Check convergence to T=0 solution
    err0 = np.linalg.norm(C0-C)
    err1 = np.linalg.norm(C1-C)
    err2 = np.linalg.norm(C2-C)
    assert err1 < err0
    assert err2 < err1
    assert err2/np.linalg.norm(C) < 1e-2

    # # test boundary case: E=1
    E = 0.5
    T = [6,12]
    C0 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0.95
    C1 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0.995
    C2 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 1
    C = dc.conc(J, T, E, t=t, model='2cxm')
    # Check convergence to T=0 solution
    err0 = np.linalg.norm(C0-C)
    err1 = np.linalg.norm(C1-C)
    err2 = np.linalg.norm(C2-C)
    assert err1 < err0
    assert err2 < err1
    assert err2/np.linalg.norm(C) < 1e-2

    # # test boundary case: E=1, Tp=0
    E = 0.5
    T = [6,12]
    C0 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0.95
    C1 = dc.conc(J, T, E, t=t, model='2cxm')
    T[0] = 0.1
    C2 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0.99
    T[0] = 0.01
    C3 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 1
    T[0] = 0
    C = dc.conc(J, T, E, t=t, model='2cxm')
    # Check convergence to T=0 solution
    err0 = np.linalg.norm(C0-C)
    err1 = np.linalg.norm(C1-C)
    err2 = np.linalg.norm(C2-C)
    err3 = np.linalg.norm(C3-C)
    assert err1 < err0
    assert err2 < err1
    assert err3 < err2
    assert err3/np.linalg.norm(C) < 1e-1

    # # test boundary case: E=1, Te=0
    E = 0.5
    T = [6,12]
    C0 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0.95
    C1 = dc.conc(J, T, E, t=t, model='2cxm')
    T[1] = 0.1
    C2 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0.99
    T[1] = 0.01
    C3 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 1
    T[1] = 0
    C = dc.conc(J, T, E, t=t, model='2cxm')
    # Check convergence to T=0 solution
    #err0 = np.linalg.norm(C0-C) # diverges before it converges
    err1 = np.linalg.norm(C1-C)
    err2 = np.linalg.norm(C2-C)
    err3 = np.linalg.norm(C3-C)
    #assert err1 < err0
    assert err2 < err1
    assert err3 < err2
    assert err3/np.linalg.norm(C) < 1e-1

    # # test boundary case: E=1, Tp=inf
    E = 0.5
    T = [6,12]
    C0 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0.95
    C1 = dc.conc(J, T, E, t=t, model='2cxm')
    T[0] = 100
    C2 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0.99
    T[0] = 1000
    C3 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 1
    T[0] = np.inf
    C = dc.conc(J, T, E, t=t, model='2cxm')
    plt.show()
    # Check convergence to T=0 solution
    #err0 = np.linalg.norm(C0-C) # diverges before convergence
    err1 = np.linalg.norm(C1-C)
    err2 = np.linalg.norm(C2-C)
    err3 = np.linalg.norm(C3-C)
    #assert err1 < err0
    assert err2 < err1
    assert err3 < err2
    assert err3/np.linalg.norm(C) < 1e-1

    # # test boundary case: E=1, Te=inf
    E = 0.5
    T = [6,12]
    C0 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0.95
    C1 = dc.conc(J, T, E, t=t, model='2cxm')
    T[1] = 100
    C2 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 0.99
    T[1] = 1000
    C3 = dc.conc(J, T, E, t=t, model='2cxm')
    E = 1
    T[1] = np.inf
    C = dc.conc(J, T, E, t=t, model='2cxm')
    # Check convergence to T=0 solution
    #err0 = np.linalg.norm(C0-C) # diverges before convergence
    err1 = np.linalg.norm(C1-C)
    err2 = np.linalg.norm(C2-C)
    err3 = np.linalg.norm(C3-C)
    #assert err1 < err0
    assert err2 < err1
    assert err3 < err2
    assert err3/np.linalg.norm(C) < 1e-1

    # Check exceptions
    E = -0.5
    T = [6,12]
    try:
        C2 = dc.conc(J, T, E, t=t, model='2cxm')
    except:
        assert True
    else:
        assert False



def test_flux_2cxm():
    # Compare against ncomp
    t = np.linspace(0, 20, 10)
    J = np.ones(len(t))
    J2c = np.stack((J, np.zeros(J.size)))

    T = [6,12]
    E = 0.1
    J0 = dc.flux(J, T, E, t=t, model='2cxm')
    E = [[1-E,1],[E,0]]
    J1 = dc.flux(J2c, T, E, t=t, model='ncomp')[0,0,:]
    assert np.linalg.norm(J1-J0)/np.linalg.norm(J1) < 1e-9

    E = 0.5
    J0 = dc.flux(J, T, E, t=t, model='2cxm')
    E = [[1-E,1],[E,0]]
    J1 = dc.flux(J2c, T, E, t=t, model='ncomp')[0,0,:]
    assert np.linalg.norm(J1-J0)/np.linalg.norm(J1) < 1e-9


    # # test boundary case: E=1
    E = 0.5
    T = [6,12]
    J0 = dc.flux(J, T, E, t=t, model='2cxm')
    E = 0.95
    J1 = dc.flux(J, T, E, t=t, model='2cxm')
    E = 0.999
    J2 = dc.flux(J, T, E, t=t, model='2cxm')
    E = 1
    J3 = dc.flux(J, T, E, t=t, model='2cxm')
    # Check convergence to solution
    err0 = np.linalg.norm(J0-J3)
    err1 = np.linalg.norm(J1-J3)
    err2 = np.linalg.norm(J2-J3)
    assert err1 < err0
    assert err2 < err1
    assert err2 < 1e-2

    # # test boundary case: Tp=0
    E = 0.5
    T = [6,12]
    J0 = dc.flux(J, T, E, t=t, model='2cxm')
    T[0] = 1
    J1 = dc.flux(J, T, E, t=t, model='2cxm')
    T[0] = 0.1
    J2 = dc.flux(J, T, E, t=t, model='2cxm')
    T[0] = 0.01
    J3 = dc.flux(J, T, E, t=t, model='2cxm')
    # Check convergence to solution
    err0 = np.linalg.norm(J0-J3)
    err1 = np.linalg.norm(J1-J3)
    err2 = np.linalg.norm(J2-J3)
    assert err1 < err0
    assert err2 < err1
    assert err2 < 1e-2

    # # test boundary case: Tp=0, Te=inf
    E = 0.5
    T = [6,12]
    J0 = dc.flux(J, T, E, t=t, model='2cxm')
    T[0] = 0.1
    T[1] = 100
    J1 = dc.flux(J, T, E, t=t, model='2cxm')
    T[0] = 0.01
    T[1] = 1000
    J2 = dc.flux(J, T, E, t=t, model='2cxm')
    T[0] = 0
    T[1] = np.inf
    J3 = dc.flux(J, T, E, t=t, model='2cxm')
    # Check convergence to solution
    err0 = np.linalg.norm(J0[1:]-J3[1:])
    err1 = np.linalg.norm(J1[1:]-J3[1:])
    err2 = np.linalg.norm(J2[1:]-J3[1:])
    assert err1 < err0
    assert err2 < err1
    assert err2 < 1e-2

    T[0] = 0.0001
    T[1] = 100
    J1 = dc.flux(J, T, E, t=t, model='2cxm')
    T[0] = 0.0
    T[1] = 100
    J2 = dc.flux(J, T, E, t=t, model='2cxm')
    assert np.linalg.norm(J1[1:]-J2[1:]) < 1e-6*np.linalg.norm(J2[1:])


def test_conc():
    # We only need to check the exceptions 
    # as all other conditions are already covered by above tests
    t = np.linspace(0, 20, 10)
    J = np.ones(len(t))
    T = [6,12]
    E = 0.1
    try:
        dc.conc(J, T, E, t=t, model='blabla') 
    except:
        assert True
    else:
        assert False  

def test_flux():
    # We only need to check the exceptions 
    # as all other conditions are already covered by above tests
    t = np.linspace(0, 20, 10)
    J = np.ones(len(t))
    T = [6,12]
    E = 0.1
    try:
        J1 = dc.flux(J, T, E, t=t, model='blabla') 
    except:
        assert True
    else:
        assert False   


if __name__ == "__main__":

    test_res_trap()
    test_prop_trap()
    test_conc_trap()
    test_flux_trap()

    test_conc_pass()
    test_flux_pass()
    test_res_pass()
    test_prop_pass()

    test_conc_comp()
    test_flux_comp()
    test_res_comp()
    test_prop_comp()

    test_conc_plug()
    test_flux_plug()
    test_res_plug()
    test_prop_plug()

    test_conc_chain()
    test_flux_chain()
    test_prop_chain()
    test_res_chain()

    test_conc_step()
    test_flux_step()
    test_prop_step()
    test_res_step()

    test_conc_free()
    test_flux_free()
    test_prop_free()
    test_res_free()

    test__K_2comp()
    test__K_ncomp()
    test__J_ncomp()
    test_conc_ncomp_prop()
    test_conc_ncomp_diag()

    test_conc_ncomp()
    test_flux_ncomp()
    test_res_ncomp()
    test_prop_ncomp()

    test_conc_2comp()
    test_flux_2comp()
    test_res_2comp()
    test_prop_2comp()

    test_conc_nscomp()
    test_flux_nscomp()

    test_conc_mmcomp()
    test_flux_mmcomp()

    test_flux_pfcomp()

    test_conc_2cxm()
    test_flux_2cxm()

    test_conc()
    test_flux()

    print('All pk tests passed!!')


import numpy as np
import dcmri as dc


def test_aif_tristan():

    t = np.arange(0, 6*60, 1)
    ca = dc.aif_tristan(t)
    assert round(max(ca), 4) == 0.0035


def test_flux_aorta():
    
    J = np.ones(100)
    Ja = dc.flux_aorta(J, FFkl=0.2)
    assert round(max(Ja), 2) == 2.87



if __name__ == "__main__":

    test_aif_tristan()
    test_flux_aorta()

    print('All pk_aorta tests passed!!')
